# %%
# # Tutorial 1 venco.py

# This tutorial showcases the general structure and workflow of venco.py, as
# well as some basic features of its 6 main classes:
# - DataParser
# - GridModeller
# - FlexEstimator
# - DiaryBuilder
# - Aggregator
# - PostProcessor

# All tutorials run on a very small subset of data from the 2017 German national
# travel survey (Mobilität in Deutschland (MiD17)), which might result in
# profiles having uncommon shapes. As such, the calculations and the examples
# proposed throughout all tutorials have the mere goal to exemplify the
# modelling steps and guide the use throughout the structure of venco.py and do
# not aim at providing an accurate quantification of demand-side flexibility
# from EVs. For a more detailed description of venco.py, you can refer to the
# documentation at https://dlr-ve.gitlab.io/esy/vencopy/vencopy/

# ## Setting up the working space

# This section allows you to import all required Python packages for data input
# and manipulation. The function os.chdir(path) allows us to point Python
# towards the top most directory which contains all useful venco.py functions
# that are going to be used in the tutorials. Additionally, we set and read in
# the input dataframe (here the MiD17) and load the necessary yaml file, which
# contains some configuration settings.

import time
from pathlib import Path

from vencopy.core.dataparsers import parse_data
from vencopy.core.gridmodellers import GridModeller
from vencopy.core.flexestimators import FlexEstimator
from vencopy.core.diarybuilders import DiaryBuilder
from vencopy.core.profileaggregators import ProfileAggregator
from vencopy.core.postprocessors import PostProcessor
from vencopy.utils.utils import load_configs, create_output_folders

start_time = time.time()

# %% We will have a look more in detail at each config file and what you can
# specify within it for each class throughtout the tutorials. For the time
# being, it is enough to know that the config files specify configurations,
# variable namings and settings for the different classes. There is one config
# file for each class, a global config and a local configuration config to
# specify eventual file paths on your machine.

base_path = Path.cwd() / "vencopy"
configs = load_configs(base_path)

# %%
# ## _DataParser_ class

# To be able to estimate EV electric consumption and flexibililty, the first
# step in the venco.py framework implies accessing a travel survey data set,
# such as the MiD. This is carried out through a parsing interface to the
# original database. In the parsing interface to the data set, three main
# operations are carried out: the read-in of the travel survey trip data, stored
# in .dta or .csv files, filtering and cleaning of the original raw data set and
# a set of variable replacement operations to allow the composition of travel
# diaries in a following step (in the DiaryBuilder class).

# In order to have consistent entry data for all variables and for different
# data sets, all database entries are harmonised, which includes generating
# unified data types and consistent variable naming. The naming convention for
# the variables and their respective input type can be specified in the venco.py
# config files that have been loaded previously.

# First off, we modify the localConfig and globalConfig files so that it point
# to the current working directory and to the database subset we will use to
# explain the different classes.

# %% Adapt relative paths in config for tutorials
configs['dev_config']['global']['relative_path']['parse_output'] = Path.cwd() / "vencopy" / configs['dev_config']['global']['relative_path']['parse_output']
configs['dev_config']['global']['relative_path']['diary_output'] = Path.cwd() / "vencopy" / configs['dev_config']['global']['relative_path']['diary_output']
configs['dev_config']['global']['relative_path']['grid_output'] = Path.cwd() / "vencopy" / configs['dev_config']['global']['relative_path']['grid_output']
configs['dev_config']['global']['relative_path']['flex_output'] = Path.cwd() / "vencopy" / configs['dev_config']['global']['relative_path']['flex_output']
configs['dev_config']['global']['relative_path']['aggregator_output'] = Path.cwd() / "vencopy" / configs['dev_config']['global']['relative_path']['aggregator_output']
configs['dev_config']['global']['relative_path']['processor_output'] = Path.cwd() / "vencopy" / configs['dev_config']['global']['relative_path']['processor_output']

# Set reference dataset
dataset_id = 'MiD17'

# Modify the localPathConfig file to point to the .csv file in the sampling
# folder in the tutorials directory where the dataset for the tutorials lies.
configs['user_config']['global']['absolute_path'][dataset_id] = Path.cwd() / 'tutorials' / 'data_sampling'

# Similarly we modify the dataset_id in the global config file
configs['dev_config']['global']['files'][dataset_id]['trips_data_raw'] = dataset_id + '.csv'

# We also modify the parseConfig by removing some of the columns that are
# normally parsed from the MiD, which are not available in our semplified test
# dataframe

del configs['dev_config']['dataparsers']['data_variables']['household_id']
del configs['dev_config']['dataparsers']['data_variables']['person_id']

create_output_folders(configs=configs)

# %% We can now run the first class and parse the dataset with the collection of
# mobility patterns into a more useful form for our scope.

data = parse_data(configs=configs)
data.process()

# %%
# ## _GridModeller_ class

# The charging infrastructure allocation makes use of a basic charging
# infrastructure model, which assumes the availability of charging stations when
# vehicles are parked. Since the analytical focus of the framework lies on a
# regional level (NUTS1-NUTS0), the infrastructure model is kept simple in the
# current version.

# Charging availability is allocated based on a binary True–False mapping to a
# respective trip purpose in the venco.py config. Thus, different scenarios
# describing different charging availability scenarios, e.g., at home or at home
# and at work etc. can be distinguished, but neither a regional differentiation
# nor a charging availability probability or distribution are assumed.

# At the end of the execution of the GridModeller class, a column representing
# the available charging power is added to the activities dataset.

grid = GridModeller(configs=configs, activities=data.activities)
grid.assign_grid()

# %%
# ## _FlexEstimator_ class

# The flexEstimator class is the final class that is used to estimate the
# charging flexibility based on driving profiles and charge connection shares.
# There are three integral inputs to the flexibililty estimation:
# - A profile describing driven distances for each vehicle
# - A profiles describing the available charging power if a vehicle is connected
#   to the grid
# - Techno–economic input assumptions

flex = FlexEstimator(configs=configs, activities=grid.activities)
flex.estimate_technical_flexibility_through_iteration()

# %%
# ## _DiaryBuilder_ class

# In the DiaryBuilder, individual trips at the survey day are consolidated into
# person-specific travel diaries comprising multiple trips.


# The daily travel diary composition consists of three main steps: reformatting
# the database, allocating trip purposes and merging the obtained dataframe with
# other relevant variables from the original database.


# In the first step, reformatting, the time dimension is transferred from the
# raw data (usually in minutes) to the necessary output format (e.g., hours).
# Each trip is split into shares, which are then assigned to the respective hour
# in which they took place, generating an hourly dataframe with a timestamp
# instead of a dataframe containing single trip entries.


# Similarly, miles driven and the trip purpose are allocated to their respective
# hour and merged into daily travel diaries. Trips are assumed to determine the
# respective person’s stay in the consecutive hours up to the next trip and
# therefore are related to the charging availability between two trips. Trip
# purposes included in surveys may comprise trips carried out for work or
# education reasons, trips returning to home, trips to shopping facilities and
# other leisure activities. Currently, trips whose purpose is not specified are
# allocated to trips returning to their own household.

diary = DiaryBuilder(configs=configs, activities=flex.activities)
diary.create_diaries()

# %%
# ## _ProfileAggregator_ class
# The ProfileAggregator class provides different methods to aggregate the output
# profiles from single vehicle level to fleet level depending on the day of the
# week.

profiles = ProfileAggregator(configs=configs, activities=diary.activities, profiles=diary)
profiles.aggregate_profiles()

# %%
# ## _PostProcessor_ class
# The last class, the PostProcessor, allows to normalise the output profiles and
# create from weekly timeseries timeseries for a whole modelling year.

post = PostProcessor(configs=configs, profiles=profiles)
post.create_annual_profiles()
post.normalise()

elapsed_time = time.time() - start_time
print(f"Elapsed time: {elapsed_time}.")

# ## Next Steps
# In the next tutorials, you will learn more in detail the internal workings of
# each class and how to customise some settings.



__maintainer__ = "Fabia Miorelli"
__license__ = "BSD-3-Clause"

import pytest
import yaml
import os

from unittest.mock import mock_open, patch
import pandas as pd
from pathlib import Path

from ...vencopy.utils.utils import load_configs, return_lowest_level_dict_keys, return_lowest_level_dict_values, replace_vec, create_output_folders, create_file_name, write_out


@pytest.fixture
def temp_dir(tmp_path):
    return tmp_path / "temp_dir"


def test_load_configs(temp_dir):
    temp_dir.mkdir(parents=True)
    (temp_dir / "config").mkdir()
    user_config_name = "user_config.yaml"
    dev_config_name = "dev_config.yaml"

    user_config_data = {"user_key": "user_value"}
    dev_config_data = {"dev_key": "dev_value"}

    with open((temp_dir / "config" / user_config_name), "w") as user_file:
        yaml.dump(user_config_data, user_file, default_flow_style=False)
    with open((temp_dir / "config" / dev_config_name), "w") as dev_file:
        yaml.dump(dev_config_data, dev_file, default_flow_style=False)

    configs = load_configs(temp_dir)

    assert "user_config" in configs
    assert "dev_config" in configs
    assert configs["user_config"] == user_config_data
    assert configs["dev_config"] == dev_config_data

    user_config_data = {}
    dev_config_data = {}
    with open((temp_dir / "config" / user_config_name), "w") as user_file:
        yaml.dump(user_config_data, user_file, default_flow_style=False)
    with open((temp_dir / "config" / dev_config_name), "w") as dev_file:
        yaml.dump(dev_config_data, dev_file, default_flow_style=False)

    configs = load_configs(temp_dir)
    assert configs == {"dev_config": {}, "user_config": {}}


def test_return_lowest_level_dict_keys():
    # Test when the input dictionary has nested dictionaries
    dictionary = {
        "key1": {
            "subkey1": "value1",
            "subkey2": {
                "subsubkey1": "value2"
            }
        },
        "key2": "value3"
    }
    expected_result = ["subsubkey1", "subkey1", "key2"]

    result = return_lowest_level_dict_keys(dictionary)
    assert set(result) == set(expected_result)

    # Test when the input dictionary has no nested dictionaries
    dictionary = {
        "key1": "value1",
        "key2": "value2",
        "key3": None
    }
    expected_result = ["key1", "key2"]

    result = return_lowest_level_dict_keys(dictionary)
    assert result == expected_result

    # Test when the input dictionary is empty
    dictionary = {}
    expected_result = []

    result = return_lowest_level_dict_keys(dictionary)
    assert result == expected_result


def test_return_lowest_level_dict_values():
    # Test when the input dictionary has nested dictionaries
    dictionary = {
        "key1": {
            "subkey1": "value1",
            "subkey2": {
                "subsubkey1": "value2"
            }
        },
        "key2": "value3"
    }
    expected_result = ["value1", "value2", "value3"]

    result = return_lowest_level_dict_values(dictionary)
    assert set(result) == set(expected_result)

    # Test when the input dictionary has no nested dictionaries
    dictionary = {
        "key1": "value1",
        "key2": "value2",
        "key3": None
    }
    expected_result = ["value1", "value2"]

    result = return_lowest_level_dict_values(dictionary)
    assert set(result) == set(expected_result)

    # Test when the input dictionary is empty
    dictionary = {}
    expected_result = []

    result = return_lowest_level_dict_values(dictionary)
    assert result == expected_result


def test_replace_vec():
    data = pd.DataFrame({
        "timestamp": [pd.to_datetime("2021-01-01 12:30:45"), pd.to_datetime("2022-02-02 13:45:00")]
    })

    # Test replacing only the year
    result = replace_vec(data["timestamp"], year=2023)
    expected_result = pd.to_datetime(["2023-01-01 12:30:45", "2023-02-02 13:45:00"])
    assert all(result == expected_result)

    # Test replacing only the month
    result = replace_vec(data["timestamp"], month=5)
    expected_result = pd.to_datetime(["2021-05-01 12:30:45", "2022-05-02 13:45:00"])
    assert all(result == expected_result)

    # Test replacing only the day
    result = replace_vec(data["timestamp"], day=15)
    expected_result = pd.to_datetime(["2021-01-15 12:30:45", "2022-02-15 13:45:00"])
    assert all(result == expected_result)

    # Test replacing only the hour
    result = replace_vec(data["timestamp"], hour=7)
    expected_result = pd.to_datetime(["2021-01-01 07:30:45", "2022-02-02 07:45:00"])
    assert all(result == expected_result)

    # Test replacing only the minute
    result = replace_vec(data["timestamp"], minute=15)
    expected_result = pd.to_datetime(["2021-01-01 12:15:45", "2022-02-02 13:15:00"])
    assert all(result == expected_result)

    # Test replacing multiple components
    result = replace_vec(data["timestamp"], year=2023, month=5, day=15, hour=7, minute=15, second=30)
    expected_result = pd.to_datetime(["2023-05-15 07:15:30", "2023-05-15 07:15:30"])
    assert all(result == expected_result)

    # Test not replacing any component
    result = replace_vec(data["timestamp"])
    expected_result = pd.to_datetime(["2021-01-01 12:30:45", "2022-02-02 13:45:00"])
    assert all(result == expected_result)


@pytest.fixture
def sample_configs(tmp_path):
    root_path = tmp_path / "sample_root"
    root_path.mkdir(parents=True)

    configs = {
        "user_config": {
            "global": {
                "absolute_path": {
                    "vencopy_root": str(root_path)
                }
            }
        }
    }

    return configs


def test_create_output_folders(sample_configs):
    root_path = Path(sample_configs["user_config"]["global"]["absolute_path"]["vencopy_root"])
    main_dir = "output"
    create_output_folders(sample_configs)

    assert os.path.exists(Path(root_path))
    assert os.path.exists(Path(root_path / main_dir))

    sub_dirs = (
        "dataparser",
        "diarybuilder",
        "gridmodeller",
        "flexestimator",
        "profileaggregator",
        "postprocessor"
    )
    for sub_dir in sub_dirs:
        assert os.path.exists(Path(root_path / main_dir / sub_dir))


def test_create_file_name():
    dev_config = {
        "global": {
            "disk_file_names": {
                "file1": "file1_dev"
            },
            "additional_label": ""
        }
    }
    user_config = {
        "global": {
            "run_label": "run123"
        }
    }

    # Test when dataset is None, manual_label is empty, and suffix is 'csv'
    result = create_file_name(dev_config=dev_config, user_config=user_config, file_name_id="file1", dataset=None, suffix="csv")
    assert result == "file1_dev__run123.csv"

    # Test when dataset is provided, manual_label is empty, and suffix is 'txt'
    result = create_file_name(dev_config=dev_config, user_config=user_config, file_name_id="file1", dataset="dataset1", suffix="txt")
    assert result == "file1_dev_dataset1_run123.txt"

    # Test when manual_label is provided, and dataset and suffix are None
    result = create_file_name(dev_config=dev_config, user_config=user_config, file_name_id="file1", dataset=None)
    assert result == "file1_dev__run123.csv"

    # Test when all parameters are provided
    result = create_file_name(dev_config=dev_config, user_config=user_config, file_name_id="file1", dataset="dataset1", suffix="txt")
    assert result == "file1_dev_dataset1_run123.txt"


def test_write_out(tmp_path):
    data = pd.DataFrame({
        "A": [1, 2, 3],
        "B": [4, 5, 6]
    })
    output_path = Path(tmp_path) / "output.csv"

    write_out(data, output_path)

    assert os.path.isfile(output_path)
    assert output_path.exists()

    loaded_data = pd.read_csv(output_path, index_col=0)
    assert data.equals(loaded_data)

    #captured = capsys.readouterr()
    #assert f"Dataset written to {output_path}." in captured.out



__maintainer__ = "Fabia Miorelli"
__license__ = "BSD-3-Clause"

import pytest

import pandas as pd

from ...vencopy.core.postprocessors import PostProcessor

# NOT TESTED: normalise(), __write_out_profiles()



__maintainer__ = "Fabia Miorelli"
__license__ = "BSD-3-Clause"


import pytest

import pandas as pd
from pathlib import Path
from dateutil import parser
from ...vencopy.core.flexestimators import FlexEstimator

@pytest.fixture
def sample_data_frame_other_filters():
    data = {
        "unique_id": [1, 1, 2, 2, 3],
        "timestamp_start": ["2023-09-01 08:00", "2023-09-01 08:45", "2023-09-01 10:00", "2023-09-01 10:30", "2023-09-01 11:00"],
        "timestamp_end": ["2023-09-01 09:00", "2023-09-01 09:30", "2023-09-01 10:15", "2023-09-01 11:15", "2023-09-01 11:30"],
        "trip_distance": [60.0, 15.0, 10.0, 10.0, 10.0],
        "travel_time": [60, 45, 90, 90, 90],
        "season": ["summer", "winter", "spring", "summer", "summer"]
    }
    data["timestamp_start"] = [parser.parse(x) for x in data["timestamp_start"]]
    data["timestamp_end"] = [parser.parse(x) for x in data["timestamp_end"]]
    return pd.DataFrame(data)

def _test_drain():

        df = sample_data_frame_other_filters()

        def _calculate_drain_by_season(x, y):
            if x == "spring" or x == "fall":
                x_new = y * 18 / 100
                return x_new
            elif x == "summer":
                x_new = (y * 18 / 100)*1.033
                return x_new
            elif x == "winter":
                x_new = (y * 18 / 100)*1.3
                return x_new
        # print(x_new)
    
        df["drain"] = df.apply(lambda row: _calculate_drain_by_season(row['season'],
                                                row['trip_distance']), axis=1)
        print(df)
        



__maintainer__ = "Fabia Miorelli"
__license__ = "BSD-3-Clause"

import pytest
import pandas as pd

from typing import Any, Literal

from ....vencopy.core.dataparsers.dataparsers import DataParser
from ....vencopy.core.dataparsers.parseMiD import ParseMiD
from ....vencopy.core.dataparsers.parseKiD import ParseKiD



__maintainer__ = "Fabia Miorelli"
__license__ = "BSD-3-Clause"

import pytest

import pandas as pd

from ...vencopy.core.diarybuilders import DiaryBuilder

# NOT TESTED: create_diaries(), __update_activities()

@pytest.fixture
def sample_configs():
    configs = {
        'user_config': {
            'global': {
                'dataset': "dataset1",
                'debug': False,
                'absolute_path': {
                    'dataset1': '/path/to/dataset1',
                    'dataset2': '/path/to/dataset2'
                    }
                },
            'diarybuilders': {
                'time_resolution': 15
            },
            'flexestimators': {
                'plugging_behaviour': {
                    "consider_plugging_behaviour": False
                }
            }
            },
        'dev_config': {
            'dataparsers': {
                'data_variables': False
                }
            }
        }
    return configs


@pytest.fixture
def sample_activities():
    data = {}
    activities = pd.DataFrame({
        "activity_id": [1, 2, 3, 4, 5],
        "activity_duration": [pd.Timedelta(minutes=76), pd.Timedelta(minutes=80), pd.Timedelta(0), pd.Timedelta(minutes=45), pd.Timedelta(minutes=1440)],
        "timestamp_start": pd.DatetimeIndex(["2023-09-12 08:00:00", "2023-09-12 10:30:00", "2023-09-12 10:30:00", "2023-09-12 10:00:00", "2023-09-12 00:00:00"]),
        "timestamp_end": pd.DatetimeIndex(["2023-09-12 09:16:00", "2023-09-12 11:50:00", "2023-09-12 10:30:00", "2023-09-12 10:45:00", "2023-09-13 00:00:00"]),
        "is_first_activity": (True, False, True, False, True),
        "is_last_activity": (False, False, False, False, True)
        })
    data['activities'] = activities
    return data


def test_diarybuilder_init(sample_configs):
    sample_activities = {}
    activities = pd.DataFrame({
        "activity_id": [1, 2, 3, 4],
        "activity_duration": [pd.Timedelta(minutes=76), pd.Timedelta(minutes=80), pd.Timedelta(0), pd.Timedelta(minutes=45)],
        "timestamp_start": pd.DatetimeIndex(["2023-09-12 08:00:00", "2023-09-12 10:30:00", "2023-09-12 10:30:00", "2023-09-12 10:00:00"]),
        "timestamp_end": pd.DatetimeIndex(["2023-09-12 09:16:00", "2023-09-12 11:50:00", "2023-09-12 10:30:00", "2023-09-12 10:45:00"])
        })
    sample_activities['activities'] = activities
    builder = DiaryBuilder(configs=sample_configs, data=sample_activities)

    assert builder.dev_config == sample_configs["dev_config"]
    assert builder.user_config == sample_configs["user_config"]
    assert builder.dataset == "dataset1"
    assert builder.data['activities'].equals(sample_activities['activities'])
    assert builder.time_resolution == 15
    assert builder.is_week_diary == False


def test_correct_timestamps(sample_configs, sample_activities):
    builder = DiaryBuilder(configs=sample_configs, data=sample_activities)
    time_resolution = 15
    result = builder._correct_timestamps(activities=sample_activities['activities'], time_resolution=time_resolution)
    expected_result = pd.DataFrame({
        "activity_duration": [pd.Timedelta(minutes=75), pd.Timedelta(minutes=75), pd.Timedelta(0), pd.Timedelta(minutes=45), pd.Timedelta(days=1)],
        "timestamp_start_corrected": pd.DatetimeIndex(["2023-09-12 08:00:00", "2023-09-12 10:30:00", "2023-09-12 10:30:00", "2023-09-12 10:00:00", "2023-09-12 00:00:00"]),
        "timestamp_end_corrected": pd.DatetimeIndex(["2023-09-12 09:15:00", "2023-09-12 11:45:00", "2023-09-12 10:30:00", "2023-09-12 10:45:00", "2023-09-13 00:00:00"]),
    })

    pd.testing.assert_series_equal(result["timestamp_start_corrected"], expected_result["timestamp_start_corrected"])
    pd.testing.assert_series_equal(result["timestamp_end_corrected"], expected_result["timestamp_end_corrected"])
    pd.testing.assert_series_equal(result["activity_duration"], expected_result["activity_duration"])


def test_removes_zero_length_activities(sample_activities):
    result = DiaryBuilder._removes_zero_length_activities(activities=sample_activities['activities'])

    assert len(result) == 4


# %% [markdown]
# # Tutorial 4 venco.py

# This tutorial aims to give a more in depth overview into the FlexEstimator
# class and showcases some features that can be customised.

from pathlib import Path
import matplotlib.pyplot as plt

from vencopy.core.dataparsers import parse_data
from vencopy.core.flexestimators import FlexEstimator
from vencopy.core.gridmodellers import GridModeller
from vencopy.core.diarybuilders import DiaryBuilder
from vencopy.core.profileaggregators import ProfileAggregator
from vencopy.utils.utils import load_configs, create_output_folders

# %%
base_path = Path.cwd() / 'vencopy'
configs = load_configs(base_path)
create_output_folders(configs=configs)

# Adapt relative paths in config for tutorials
configs['dev_config']['global']['relative_path']['parse_output'] = Path.cwd() / "vencopy" / configs['dev_config']['global']['relative_path']['parse_output']
configs['dev_config']['global']['relative_path']['diary_output'] = Path.cwd() / "vencopy" / configs['dev_config']['global']['relative_path']['diary_output']
configs['dev_config']['global']['relative_path']['grid_output'] = Path.cwd() / "vencopy" / configs['dev_config']['global']['relative_path']['grid_output']
configs['dev_config']['global']['relative_path']['flex_output'] = Path.cwd() / "vencopy" / configs['dev_config']['global']['relative_path']['flex_output']
configs['dev_config']['global']['relative_path']['aggregator_output'] = Path.cwd() / "vencopy" / configs['dev_config']['global']['relative_path']['aggregator_output']
configs['dev_config']['global']['relative_path']['processor_output'] = Path.cwd() / "vencopy" / configs['dev_config']['global']['relative_path']['processor_output']

# Set reference dataset
dataset_id = 'MiD17'

# Modify the localPathConfig file to point to the .csv file in the sampling
# folder in the tutorials directory where the dataset for the tutorials lies.
configs["user_config"]["global"]["absolute_path"]["vencopy_root"] = Path.cwd()
configs['user_config']['global']['absolute_path'][dataset_id] = Path.cwd() / 'tutorials' / 'data_sampling'

# Similarly we modify the dataset_id in the global config file
configs['dev_config']['global']['files'][dataset_id]['trips_data_raw'] = dataset_id + '.csv'


# We also modify the parseConfig by removing some of the columns that are
# normally parsed from the MiD, which are not available in our semplified test
# dataframe
del configs['dev_config']['dataparsers']['data_variables']['household_id']
del configs['dev_config']['dataparsers']['data_variables']['person_id']


# %%
# ## FlexEstimator config file

# The FlexEstimator config file contains the technical specifications.

configs['user_config']['flexestimators']

# %%
# ## _FlexEstimator_ class

# To use the FlexEstimator class, we first need to run the DataParses as well as
# the GridModeller as they have an argument to the FlexEstimator class.


base_path = Path.cwd() / 'vencopy'

data = parse_data(configs=configs)
data = data.process()

grid = GridModeller(configs=configs, activities=data)
grid.assign_grid()


# %%
# Now we can display results on the distance all the trips have

# Filter out rows where trip_distance is not NaN
filtered_df = data.dropna(subset=['trip_distance'])

max_index = filtered_df['index'].max()

# Plot trip distances
plt.figure(figsize=(10, 6))
plt.scatter(filtered_df.index, filtered_df['trip_distance'], color='blue', alpha=0.5)
plt.title('Trip Distance Distribution')
plt.xlabel('Index')
plt.ylabel('Trip Distance')
plt.grid(True)
plt.xlim(0, max_index)
plt.show()

# %% Estimate charging flexibility based on driving profiles and charge
# connection
flex = FlexEstimator(configs=configs, activities=grid.activities)
flex.estimate_technical_flexibility_through_iteration()


# %%
diary = DiaryBuilder(configs=configs, activities=flex.activities)
diary.create_diaries()

profiles = ProfileAggregator(configs=configs, activities=diary.activities, profiles=diary)
profiles.aggregate_profiles()

plt.figure(figsize=(10, 6))
plt.plot(profiles.uncontrolled_charging_weekly.index, profiles.uncontrolled_charging_weekly.iloc[:])
plt.xlim(profiles.uncontrolled_charging_weekly.index[0], profiles.uncontrolled_charging_weekly.index[-1])
plt.title('Uncontrolled Charge Volume')
plt.xlabel('Time')
plt.ylabel('Charge Volume')
plt.grid(True)
plt.show()


# %%
# To analyse its influence on the demand-side flexibility from EV, we will
# though charge the assumed size of the battery from 50 kWh to 100 kWh.

configs['user_config']['flexestimators']['battery_capacity'] = 20.0

# %% Estimate charging flexibility based on driving profiles and charge
# connection
flex = FlexEstimator(configs=configs, activities=grid.activities)
flex.estimate_technical_flexibility_through_iteration()

diary = DiaryBuilder(configs=configs, activities=flex.activities)
diary.create_diaries()

profiles = ProfileAggregator(configs=configs, activities=diary.activities, profiles=diary)
profiles.aggregate_profiles()

plt.figure(figsize=(10, 6))
plt.plot(profiles.uncontrolled_charging_weekly.index, profiles.uncontrolled_charging_weekly.iloc[:])
plt.xlim(profiles.uncontrolled_charging_weekly.index[0], profiles.uncontrolled_charging_weekly.index[-1])
plt.title('Uncontrolled Charge Volume')
plt.xlabel('Time')
plt.ylabel('Charge Volume')
plt.grid(True)
plt.show()

# By reducing the battery capacity to 20 kWh we can see from the graphs flatter
# curves, which results from vehicles charging less and reduced range of
# profiles that can be considered.

# ## Next Steps
# Come back, there will be more upcoming tutorials! :)



__maintainer__ = "Niklas Wulff, Fabia Miorelli"
__license__ = "BSD-3-Clause"


import time
import warnings
import logging

from pathlib import Path

from vencopy.core.dataparsers import parse_data
from vencopy.core.gridmodellers import GridModeller
from vencopy.core.flexestimators import FlexEstimator
from vencopy.core.diarybuilders import DiaryBuilder
from vencopy.core.profileaggregators import ProfileAggregator
from vencopy.core.postprocessors import PostProcessor
from vencopy.utils.utils import load_configs, create_output_folders

warnings.simplefilter(action='ignore', category=FutureWarning)
logging.basicConfig(level=logging.INFO, filename='vencopy.log', filemode='w', format="{%(filename)s: %(lineno)d} %(message)s")

if __name__ == "__main__":
    start_time = time.time()

    base_path = Path(__file__).parent / "vencopy"
    configs = load_configs(base_path=base_path)
    create_output_folders(configs=configs)

    data = parse_data(configs=configs)
    data.process()

    grid = GridModeller(configs=configs, data=data.data)
    grid.assign_grid()

    flex = FlexEstimator(configs=configs, data=grid.data)
    flex.estimate_technical_flexibility_through_iteration()

    diary = DiaryBuilder(configs=configs, data=flex.data)
    diary.create_diaries()

    profiles = ProfileAggregator(configs=configs, data=diary.data, profiles=diary)
    profiles.aggregate_profiles()

    post = PostProcessor(configs=configs, profiles=profiles, data=diary.data)
    post.normalise_profiles()
    post.create_annual_profiles()
    post.generate_fleet_profiles()

    elapsed_time = time.time() - start_time
    logging.info(f"Elapsed time: {elapsed_time}.")



__maintainer__ = "Niklas Wulff, Fabia Miorelli"
__license__ = "BSD-3-Clause"

import logging

import pandas as pd

from ...core.dataparsers.dataparsers import IntermediateParsing
from ...core.dataparsers.parkinference import ParkInference


class ParseMiD(IntermediateParsing):
    def __init__(self, configs: dict, dataset: str):
        """
        Inherited data class to differentiate between abstract interfaces such
        as vencopy internal variable namings and data set specific functions
        such as filters. Specific class for the German MiD B2 dataset.

        Args:
            configs (dict): A dictionary containing a user_config dictionary and a dev_config dictionary.
            dataset (str): Abbreviation of the National Travel Survey to be parsed.
        """
        super().__init__(configs=configs, dataset=dataset)
        self.park_inference = ParkInference(configs=configs)

    def _harmonise_variables(self):
        """
        Harmonizes the input data variables to match internal venco.py names
        given as specified in the mapping in parseConfig['data_variables']. So
        far mappings for MiD08 and MiD17 are given. Since the MiD08 does not
        provide a combined household and person unique identifier, it is
        synthesized of the both IDs.
        """
        replacement_dict = self._create_replacement_dict(self.dataset, self.dev_config["dataparsers"]["data_variables"])
        activities_renamed = self.trips.rename(columns=replacement_dict)
        if self.dataset == "MiD08":
            activities_renamed["household_person_id"] = (
                activities_renamed["household_id"].astype("string") + activities_renamed["person_id"].astype("string")
            ).astype("int")
        self.trips = activities_renamed
        logging.info("Finished harmonisation of variables.")

    def __add_string_columns(self, weekday=True, purpose=True):
        """
        Adds string columns for either weekday or purpose.

        Args:
            weekday (bool, optional): Boolean identifier if weekday should be
            added in a separate column as string. Defaults to True. purpose
            (bool, optional): Boolean identifier if purpose should be added in a
            separate column as string. Defaults to True.
        """
        if weekday:
            self._add_string_column_from_variable(col_name="weekday_string", var_name="trip_start_weekday")
        if purpose:
            self._add_string_column_from_variable(col_name="purpose_string", var_name="trip_purpose")

    @staticmethod
    def _cleanup_dataset(activities):
        activities.drop(
            columns=['is_driver',
                     'person_id',
                     'household_person_id',
                     'trip_scale_factor',
                     'trip_end_next_day',
                     'trip_is_intermodal',
                     'trip_purpose',
                     'weekday_string',
                     'is_first_trip'], inplace=True)
        return activities

    def process(self) -> pd.DataFrame:
        """
        Wrapper function for harmonising and filtering the trips dataset as well
        as adding parking rows.
        """
        self._load_data()
        self._select_columns()
        self._harmonise_variables()
        self._harmonise_variables_unique_id_names()
        self._convert_types()
        self.__add_string_columns()
        self._compose_start_and_end_timestamps()
        self._update_end_timestamp(trips=self.trips)
        self._assign_season(self.trips)
        if self.user_config["dataparsers"]["seasons"]["filter_season_post_assignment"]:
            self.filter_for_season(self.trips)
        self._check_filter_dict(dictionary=self.filters)
        self._filter(filters=self.filters)
        self._extract_df_debug_mode()
        self._subset_area_type()
        self.data = self.park_inference.add_parking_rows(trips=self.trips)
        self.data['activities'] = self._cleanup_dataset(activities=self.data['activities'])
        self.write_output()
        logging.info("Parsing MiD dataset completed.")
        return self.data



__maintainer__ = "Niklas Wulff, Fabia Miorelli"
__license__ = "BSD-3-Clause"


import logging
import warnings
from pathlib import Path
from zipfile import ZipFile

import pandas as pd

from ...utils.metadata import read_metadata_config, write_out_metadata
from ...utils.utils import (
    create_file_name,
    fall_start,
    return_lowest_level_dict_keys,
    return_lowest_level_dict_values,
    spring_start,
    summer_start,
    winter_start,
    write_out,
)


class DataParser:
    def __init__(self, configs: dict, dataset: str):
        """
        Basic class for parsing a mobility survey trip data set. Currently both
        German travel surveys MiD 2008 and MiD 2017 are pre-configured and one
        of the two can be given (default: MiD 2017).
        The data set can be provided from an encrypted file on a server in
        which case the link to the ZIP-file as well as a link to the file
        within the ZIP-file have to be supplied in the globalConfig and a
        password has to be supplied in the parseConfig.
        Columns relevant for the EV simulation are selected from the entirety
        of the data and renamed to venco.py internal variable names given in
        the dictionary parseConfig['data_variables'] for the respective survey
        data set. Manually configured exclude, include, greater_than and
        smaller_than filters are applied as they are specified in parseConfig.
        For some columns, raw data is transferred to human readable strings
        and respective columns are added. Pandas timestamp columns are
        synthesized from the given trip start and trip end time information.

        Args:
            configs (dict): A dictionary containing multiple yaml config files
            dataset (str): Abbreviation representation the National Travel
            Survey to be parsed
        """
        self.user_config = configs["user_config"]
        self.dev_config = configs["dev_config"]
        self.debug = configs["user_config"]["global"]["debug"]
        self.season = self.user_config["global"]["consider_temperature_cycle_dependency"]["season"]
        self.dataset = self._check_dataset_id(dataset=dataset)
        self.data = {}
        self.raw_data_path = (
            Path(self.user_config["global"]["absolute_path"][self.dataset])
            / self.dev_config["global"]["files"][self.dataset]["trips_data_raw"]
        )
        self.raw_data = None
        self.trips = None
        self.activities = None
        self.filters = {}
        self.number_lines_debug = self.user_config["global"]["number_lines_debug"]
        logging.info("Generic file parsing properties set up.")

    def _load_data(self):
        """
        Checks the load_encrypted value and then loads the data specified in
        self.raw_data_path and stores it in self.raw_data. Checks the
        number_of_debug value only takes on those rows in the self.raw_data.

        """
        load_encrypted = False
        if load_encrypted:
            logging.info(
                f"Starting to retrieve encrypted data file from {self.raw_data_path}."
            )
            self._load_encrypted_data(
                zip_path=self.raw_data_path, path_zip_data=self.raw_data_path
            )
        else:
            logging.info(f"Starting to retrieve local data file from {self.raw_data_path}.")
            self._load_unencrypted_data()

    def _extract_df_debug_mode(self):
        if self.debug:
            logging.info("Running in debug mode.")
            if self.user_config['dataparsers']['subset_vehicle_segment']:
                self.trips = self.trips.sort_values(['trip_start_weekday', 'season', 'vehicle_segment_string']).reset_index(drop=True)
                number_lines_per_cluster = round(self.number_lines_debug / (len(self.trips.trip_start_weekday.unique()) * len(self.trips.season.unique()) * len(self.trips.vehicle_segment_string.unique())))
                self.trips = self.trips.groupby(['trip_start_weekday', 'season', 'vehicle_segment_string']).head(number_lines_per_cluster).reset_index(drop=True)
            else:
                self.trips = self.trips.sort_values(['trip_start_weekday', 'season']).reset_index(drop=True)
                number_lines_per_cluster = round(self.number_lines_debug / (len(self.trips.trip_start_weekday.unique()) * len(self.trips.season.unique())))
                self.trips = self.trips.groupby(['trip_start_weekday', 'season']).head(number_lines_per_cluster).reset_index(drop=True)
        return self.trips

    def _load_unencrypted_data(self) -> pd.DataFrame:
        """
        Loads data specified in self.raw_data_path and stores it in
        self.raw_data. Raises an exception if a invalid suffix is specified in
        self.raw_data_path.

        Returns:
            pd.DataFrame: raw_data
        """
        if self.raw_data_path.suffix == ".dta":
            self.raw_data = pd.read_stata(
                self.raw_data_path,
                convert_categoricals=False,
                convert_dates=False,
                preserve_dtypes=False,
            )
        elif self.raw_data_path.suffix == ".csv":
            self.raw_data = pd.read_csv(self.raw_data_path)
        else:
            Exception(
                f"Data type {self.raw_data_path.suffix} not yet specified. Available types so far are .dta and .csv"
            )
        logging.info(
            f"Finished loading {len(self.raw_data)} rows of raw data of type {self.raw_data_path.suffix}."
        )
        return self.raw_data

    def _load_encrypted_data(self, zip_path, path_zip_data):
        """
        Function to provide the possibility to access encrypted zip files.
        An encryption password has to be given in user_config.yaml in order to access the encrypted file.
        Loaded data is stored in self.raw_data.

        Args:
            zip_path : Path from current working directory to the zip file or absolute path to zipfile
            path_zip_data : Path to trip data file within the encrypted zipfile
        """
        with ZipFile(zip_path) as myzip:
            if ".dta" in path_zip_data:
                self.raw_data = pd.read_stata(
                    myzip.open(
                        path_zip_data,
                        pwd=bytes(
                            self.user_config["dataparsers"]["encryption_password"],
                            encoding="utf-8",
                        ),
                    ),
                    convert_categoricals=False,
                    convert_dates=False,
                    preserve_dtypes=False,
                )
            else:  # if '.csv' in path_zip_data:
                self.raw_data = pd.read_csv(
                    myzip.open(
                        path_zip_data,
                        pwd=bytes(
                            self.user_config["dataparsers"]["encryption_password"],
                            encoding="utf-8",
                        ),
                    ),
                    sep=";",
                    decimal=",",
                )

        logging.info(
            f"Finished loading {len(self.raw_data)} rows of raw data of type {self.raw_data_path.suffix}."
        )

    def _check_dataset_id(self, dataset: str) -> str:
        """
        General check if data set ID is defined in dev_config.yaml

        Args:
            dataset (str): String declaring all possible dataset

        Returns:
            str: String with the dataset name
        """
        available_dataset_ids = self.dev_config["dataparsers"]["data_variables"][
            "dataset"
        ]
        assert dataset in available_dataset_ids, (
            f"Defined dataset {dataset} not specified "
            f"under data_variables in dev_config. "
            f"Specified dataset_ids are {available_dataset_ids}."
        )
        return dataset

    def _harmonise_variables(self):
        """
        Harmonizes the input data variables to match internal venco.py names
        given as specified in the mapping in dev_config['data_variables'].
        Since the MiD08 does not provide a combined household and person
        unique identifier in the child class, it is synthesized of the both IDs.
        """
        replacement_dict = self._create_replacement_dict(
            self.dataset, self.dev_config["dataparsers"]["data_variables"]
        )
        data_renamed = self.trips.rename(columns=replacement_dict)
        self.trips = data_renamed
        logging.info("Finished harmonisation of variables.")

    @staticmethod
    def _create_replacement_dict(dataset: str, data_variables: dict) -> dict:
        """
        Creates the mapping dictionary from raw data variable names to venco.py
        internal variable names as specified in dev_config.yaml
        for the specified dataset.

        Args:
            dataset (str): A list of strings declaring the dataset_id to be read
            data_variables (dict): data_variables contain the column names from the dataset

        Raises:
            ValueError: Raises a value error if the dataset is not specified in the dev_config

        Returns:
            dict: A dictionary with internal names as keys and raw data column names as values
        """
        if dataset not in data_variables["dataset"]:
            raise ValueError(
                f"Dataset {dataset} not specified in dev_config variable dictionary."
            )
        list_index = data_variables["dataset"].index(dataset)
        return {val[list_index]: key for (key, val) in data_variables.items()}

    @staticmethod
    def _check_filter_dict(dictionary):
        """
        Checking if all values of filter dictionaries are of type list.
        Currently only checking if list of list str not typechecked
        all(map(self.__checkStr, val). Conditionally triggers an assert.

        Args:
            dictionary : Dictionary containing filters in the dev_config dictionary
        """
        assert all(
            isinstance(val, list) for val in return_lowest_level_dict_values(dictionary)
        ), "Not all values in filter dictionaries are lists."

    def _filter(self, filters: dict = None):
        """
        Wrapper function to carry out filtering for the four filter logics of
        including, excluding, greater_than and smaller_than.
        If a filters is defined with a different key, a warning is thrown.
        Filters are defined inclusively, thus boolean vectors will select
        elements (TRUE) that stay in the data set. The function operates on self.trips class-internally.

        Args:
            filters (dict, optional): Defaults to None
        """
        logging.info(
            f"Starting filtering, applying {len(return_lowest_level_dict_keys(filters))} filters."
        )

        # Application of simple value-based filters
        simple_filters = self._simple_filters()
        self.data_simple = self.trips[simple_filters.all(axis="columns")]

        # Application of sophisticated filters
        complex_filters = self._complex_filters(data=self.data_simple)
        self.trips = self.data_simple.loc[complex_filters.all(axis="columns"), :]

        # Print user feedback on filtering
        self._filter_analysis(simple_filters.join(complex_filters))

    def _simple_filters(self) -> pd.DataFrame:
        """
        Apply single-column scalar value filtering as defined in the config.

        Returns:
            pd.DataFrame: DataFrame with boolean columns for include, exclude, greater_than and smaller_than filters, True
            means keep the row
        """
        simple_filter = pd.DataFrame(index=self.trips.index)

        # Simple filters checking single columns for specified values
        for i_key, i_value in self.filters.items():
            if i_key == "include" and i_value:
                simple_filter = simple_filter.join(
                    self._set_include_filter(
                        dataset=self.trips, include_filter_dict=i_value
                    )
                )
            elif i_key == "exclude" and i_value:
                simple_filter = simple_filter.join(
                    self._set_exclude_filter(
                        dataset=self.trips, exclude_filter_dict=i_value
                    )
                )
            elif i_key == "greater_than" and i_value:
                simple_filter = simple_filter.join(
                    self._set_greater_than_filter(
                        dataset=self.trips, greater_than_filter_dict=i_value
                    )
                )
            elif i_key == "smaller_than" and i_value:
                simple_filter = simple_filter.join(
                    self._set_smaller_than_filter(
                        dataset=self.trips, smaller_than_filter_dict=i_value
                    )
                )
            elif i_key not in ["include", "exclude", "greater_than", "smaller_than"]:
                warnings.warn(
                    f"A filter dictionary was defined in the dev_config with an unknown filtering key."
                    f"Current filtering keys comprise include, exclude, smaller_than and greater_than."
                    f"Continuing with ignoring the dictionary {i_key}"
                )
        return simple_filter

    @staticmethod
    def _set_include_filter(
        dataset: pd.DataFrame, include_filter_dict: dict
    ) -> pd.DataFrame:
        """
        Read-in function for include filter dict from dev_config.yaml

        Args:
            dataset (pd.DataFrame): dataframe on which the include filter can be added
            include_filter_dict (dict): Dictionary of include filters defined in dev_config.yaml

        Returns:
            pd.DataFrame: Dataframe including the variables specified
        """
        inc_filter_cols = pd.DataFrame(
            index=dataset.index, columns=include_filter_dict.keys()
        )
        for inc_col, inc_elements in include_filter_dict.items():
            inc_filter_cols[inc_col] = dataset[inc_col].isin(inc_elements)
        return inc_filter_cols

    @staticmethod
    def _set_exclude_filter(
        dataset: pd.DataFrame, exclude_filter_dict: dict
    ) -> pd.DataFrame:
        """
        Read-in function for exclude filter dict from dev_config.yaml

        Args:
            dataset (pd.DataFrame): dataframe on which the exclude filter can be added
            exclude_filter_dict (dict): Dictionary of exclude filters defined in dev_config.yaml

        Returns:
            pd.DataFrame: Filtered dataframe
        """
        excl_filter_cols = pd.DataFrame(
            index=dataset.index, columns=exclude_filter_dict.keys()
        )
        for exc_col, exc_elements in exclude_filter_dict.items():
            excl_filter_cols[exc_col] = ~dataset[exc_col].isin(exc_elements)
        return excl_filter_cols

    @staticmethod
    def _set_greater_than_filter(dataset: pd.DataFrame, greater_than_filter_dict: dict):
        """
        Read-in function for greater_than filter dict from dev_config.yaml

        Args:
            dataset (pd.DataFrame): dataframe on which the greater than filter can be added
            greater_than_filter_dict (dict): Dictionary of greater than filters defined in dev_config.yaml

        Returns:
            _type_: Filtered dataframe
        """
        greater_than_filter_cols = pd.DataFrame(
            index=dataset.index, columns=greater_than_filter_dict.keys()
        )
        for greater_col, greater_elements in greater_than_filter_dict.items():
            greater_than_filter_cols[greater_col] = (
                dataset[greater_col] >= greater_elements.pop()
            )
            if len(greater_elements) > 0:
                warnings.warn(
                    f"You specified more than one value as lower limit for filtering column {greater_col}."
                    f"Only considering the last element given in the dev_config."
                )
        return greater_than_filter_cols

    @staticmethod
    def _set_smaller_than_filter(
        dataset: pd.DataFrame, smaller_than_filter_dict: dict
    ) -> pd.DataFrame:
        """
        Read-in function for smaller_than filter dict from dev_config.yaml

        Args:
            dataset (pd.DataFrame): dataframe on which the smaller_than filter can be added
            smaller_than_filter_dict (dict): Dictionary of smaller than filters defined in dev_config.yaml

        Returns:
            pd.DataFrame: Filtered dataframe
        """
        smaller_than_filter_cols = pd.DataFrame(
            index=dataset.index, columns=smaller_than_filter_dict.keys()
        )
        for smaller_col, smaller_elements in smaller_than_filter_dict.items():
            smaller_than_filter_cols[smaller_col] = (
                dataset[smaller_col] <= smaller_elements.pop()
            )
            if len(smaller_elements) > 0:
                warnings.warn(
                    f"You specified more than one value as upper limit for filtering column {smaller_col}."
                    f"Only considering the last element given in the dev_config."
                )
        return smaller_than_filter_cols

    def _complex_filters(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Collects filters that compare multiple columns or derived variables or calculation results thereof. True
        in this filter means "keep row". The function needs self.trips to determine the length and the index of the
        return argument.

        Returns:
            pd.DataFrame: DataFrame with a boolean column per complex filter, 'True' means keep the row in the trips
            data set
        """
        complex_filters = pd.DataFrame(index=data.index)
        lower_speed_threshold = self.dev_config["dataparsers"]["filters"][
            "lower_speed_threshold"
        ]
        higher_speed_threshold = self.dev_config["dataparsers"]["filters"][
            "higher_speed_threshold"
        ]
        complex_filters = complex_filters.join(
            self._filter_inconsistent_speeds(
                dataset=data,
                lower_speed_threshold=lower_speed_threshold,
                higher_speed_threshold=higher_speed_threshold,
            )
        )
        complex_filters = complex_filters.join(
            self._filter_inconsistent_travel_times(dataset=data)
        )
        complex_filters = complex_filters.join(
            self._filter_overlapping_trips(dataset=data)
        )
        return complex_filters

    @staticmethod
    def _filter_inconsistent_speeds(
        dataset: pd.DataFrame, lower_speed_threshold, higher_speed_threshold
    ) -> pd.Series:
        """
        Filter out trips with inconsistent average speed. These trips are mainly trips where survey participant
        responses suggest that participants were travelling for the entire time they took for the whole purpose
        (driving and parking) and not just for the real travel.

        Args:
            dataset (pd.DataFrame): dataframe on which the inconsistent speed filter can be added
            lower_speed_threshold : Dictionary of lower speed threshold filters defined in dev_config.yaml
            higher_speed_threshold : Dictionary of higher speed threshold filters defined in dev_config.yaml

        Returns:
            pd.Series: Boolean vector with observations marked True that should be kept in the dataset
        """
        dataset["average_speed"] = dataset["trip_distance"] / (
            dataset["travel_time"] / 60
        )
        dataset = (dataset["average_speed"] > lower_speed_threshold) & (
            dataset["average_speed"] <= higher_speed_threshold
        )
        return dataset

    @staticmethod
    def _filter_inconsistent_travel_times(dataset_in: pd.DataFrame) -> pd.Series:
        """
        Calculates a travel time from the given timestamps and compares it
        to the travel time given by the interviewees. Selects observations where
        timestamps are consistent with the travel time given.

        Args:
            dataset_in (pd.DataFrame): dataframe on which the inconsistent travel time filter can be added

        Returns:
            pd.Series: Boolean vector with observations marked True that should be kept in the dataset
        """
        dataset = dataset_in.copy()
        dataset["travel_time_ts"] = (
            (dataset["timestamp_end"] - dataset["timestamp_start"])
            .dt.total_seconds()
            .div(60)
            .astype(int)
        )
        filt = dataset["travel_time_ts"] == dataset["travel_time"]
        filt.name = "travel_time"
        return filt

    @staticmethod
    def _filter_overlapping_trips(dataset, lookahead_periods: int = 7) -> pd.DataFrame:
        """
        Filter out trips carried out by the same car as next (second next, third next up to period next etc) trip but
        overlap with at least one of the period next trips.

        Args:
            data (pd.DataFrame): Trip data set including the two variables timestamp_start and timestamp_end
            characterizing a trip

        Returns:
            Pandas DataFrame containing periods columns comparing each trip to their following trips. If True, the
            trip does not overlap with the trip following after period trips (e.g. period==1 signifies no overlap with
            next trip, period==2 no overlap with second next trip etc.)
        """
        lst = []
        for profile in range(1, lookahead_periods + 1):
            ser = DataParser._identify_overlapping_trips(dataset, period=profile)
            ser.name = f"profile={profile}"
            lst.append(ser)
        ret = pd.concat(lst, axis=1).all(axis=1)
        ret.name = "trip_ends_after_next_start"

        d = dataset.copy()
        d["overlap"] = ret
        filter_ids = d[["unique_id", "overlap"]].groupby(by="unique_id").all()
        d["index"] = d.index
        d = d.set_index("unique_id", drop=False)
        d["overlap"] = filter_ids
        d = d.set_index("index")
        return d["overlap"]

    @staticmethod
    def _identify_overlapping_trips(dataset_in: pd.DataFrame, period: int) -> pd.Series:
        """
        Calculates a boolean vector of same length as dat that is True if the current trip does not overlap with
        the next trip. "Next" can relate to the consecutive trip (if period==1) or to a later trip defined by the
        period (e.g. for period==2 the trip after next). For determining if an overlap occurs, the end timestamp of the
        current trip is compared to the start timestamp of the "next" trip.

        Args:
            dataset (pd.DataFrame): A trip data set containing consecutive trips containing at least the columns id_col,
                timestamp_start, timestamp_end
            period (int): Forward looking period to compare trip overlap. Should be the maximum number of trip that one
                vehicle carries out in a time interval (e.g. day) in the data set

        Returns:
            pd.Series: A boolean vector that is True if the trip does not overlap with the period-next trip but belongs
                to the same vehicle
        """
        dataset = dataset_in.copy()
        dataset["is_same_id_as_previous"] = dataset["unique_id"] == dataset[
            "unique_id"
        ].shift(period)
        dataset["trip_starts_after_previous_trip"] = dataset[
            "timestamp_start"
        ] >= dataset["timestamp_end"].shift(period)
        return ~(
            dataset["is_same_id_as_previous"]
            & ~dataset["trip_starts_after_previous_trip"]
        )

    @staticmethod
    def _filter_analysis(filter_data: pd.DataFrame):
        """
                Function returns the total number of remaining rows after applying the filters.

                Args:
                    filter_data (pd.DataFrame): dataframe after all filters have been applied
        """
        len_data = sum(filter_data.all(axis="columns"))
        # bool_dict = {i_column: sum(filter_data[i_column]) for i_column in filter_data}
        # logging.info("The following number of observations were taken into account after filtering:")
        # pprint.pprint(bool_dict)
        # logging.info(f'{filter_data["averageSpeed"].sum()} trips have plausible average speeds')
        # logging.info(f'{(~filter_data["tripDoesNotOverlap"]).sum()} trips overlap and were thus filtered out')
        logging.info(
            f"All filters combined yielded that a total of {len_data} trips are taken into account."
        )
        logging.info(
            f"This corresponds to {len_data / len(filter_data)* 100} percent of the original data."
        )

    def process(self):
        """
        Wrapper function for harmonising and filtering the dataset.
        """
        raise NotImplementedError("A process method for DataParser is not implemented.")

    def write_output(self):
        """
        This saves the output from the dataparser class to a csv file.
        """
        if self.user_config["global"]["write_output_to_disk"]["parse_output"]:
            root = Path(self.user_config["global"]["absolute_path"]["vencopy_root"])
            folder = self.dev_config["global"]["relative_path"]["parse_output"]
            file_name = create_file_name(
                dev_config=self.dev_config,
                user_config=self.user_config,
                file_name_id="output_dataparser",
                dataset=self.dataset,
            )
            write_out(data=self.activities, path=root / folder / file_name)
            #self._write_metadata(file_name=root / folder / file_name)

    def generate_metadata(self, metadata_config, file_name):
        metadata_config["name"] = file_name
        metadata_config["title"] = "National Travel Survey activities dataframe"
        metadata_config["description"] = "Trips and parking activities from venco.py"
        metadata_config["sources"] = [
            f for f in metadata_config["sources"] if f["title"] in self.dataset
        ]
        reference_resource = metadata_config["resources"][0]
        this_resource = reference_resource.copy()
        this_resource["name"] = file_name.rstrip(".csv")
        this_resource["path"] = file_name
        these_fields = [
            f
            for f in reference_resource["schema"][self.dataset]["fields"]["dataparsers"]
            if f["name"] in self.activities.columns
        ]
        this_resource["schema"] = {"fields": these_fields}
        metadata_config["resources"].pop()
        metadata_config["resources"].append(this_resource)
        return metadata_config

    def _write_metadata(self, file_name):
        metadata_config = read_metadata_config()
        class_metadata = self.generate_metadata(
            metadata_config=metadata_config, file_name=file_name.name
        )
        write_out_metadata(
            metadata_yaml=class_metadata,
            file_name=file_name.as_posix().replace(".csv", ".metadata.yaml"),
        )


class IntermediateParsing(DataParser):
    def __init__(self, configs: dict, dataset: str):
        """
        Intermediate parsing class.

        Args:
            configs (dict): venco.py config dictionary consisting at least of the config dictionaries
            dataset (str): dataset to be processed
        """
        super().__init__(configs, dataset=dataset)
        self.filters = self.dev_config["dataparsers"]["filters"][self.dataset]
        self.var_datatype_dict = {}
        self.columns = self._compile_variable_list()

    def _compile_variable_list(self) -> list:
        """
        Clean up the replacement dictionary of raw data file variable (column)
        names. This has to be done because some variables that may be relevant
        for the analysis later on are only contained in one raw data set while
        not contained in another one. E.g. if a trip is an intermodal trip was
        only assessed in the MiD 2017 while it was not in the MiD 2008.
        This has to be mirrored by the filter dict for the respective dataset.

        Returns:
            list: List of variables
        """
        list_index = self.dev_config["dataparsers"]["data_variables"]["dataset"].index(
            self.dataset
        )
        variables = [
            val[list_index] if val[list_index] != "NA" else "NA"
            for _, val in self.dev_config["dataparsers"]["data_variables"].items()
        ]

        variables.remove(self.dataset)
        self._remove_na(variables)
        return variables

    @staticmethod
    def _remove_na(variables: list):
        """
        Removes all strings that can be capitalized to 'NA' from the list
        of variables.

        Args:
            variables (list): List of variables of the mobility dataset
        """
        ivars = [i_variable.upper() for i_variable in variables]
        counter = 0
        for indeces, i_variable in enumerate(ivars):
            if i_variable == "NA":
                del variables[indeces - counter]
                counter += 1

    def _select_columns(self):
        """
        Function to filter the raw_data for only relevant columns as specified
        by parseConfig and cleaned in self.compileVariablesList().
        Stores the subset of data in self.trips
        """
        self.trips = self.raw_data.loc[:, self.columns]

    def _convert_types(self):
        """
        Convert raw column types to predefined python types as specified in
        parseConfig['input_data_types'][dataset]. This is mainly done for
        performance reasons. But also in order to avoid index values that are
        of type int to be cast to float. The function operates only on
        self.trips and writes back changes to self.trips
        """
        conversion_dict = self.dev_config["dataparsers"]["input_data_types"][
            self.dataset
        ]
        keys = {
            i_column
            for i_column in conversion_dict.keys()
            if i_column in self.trips.columns
        }
        self.var_datatype_dict = {
            key: conversion_dict[key] for key in conversion_dict.keys() & keys
        }
        self.trips = self.trips.astype(self.var_datatype_dict)

    def _complex_filters(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Collects filters that compare multiple columns or derived variables or calculation results thereof. True
        in this filter means "keep row". The function needs self.trips to determine the length and the index of the
        return argument.

        Returns:
            pd.DataFrame: DataFrame with a boolean column per complex filter. True means keep the row in the activities
            data set
        """
        complex_filters = pd.DataFrame(index=data.index)
        lower_speed_threshold = self.dev_config["dataparsers"]["filters"][
            "lower_speed_threshold"
        ]
        higher_speed_threshold = self.dev_config["dataparsers"]["filters"][
            "higher_speed_threshold"
        ]
        complex_filters = complex_filters.join(
            self._filter_inconsistent_speeds(
                dataset=self.trips,
                lower_speed_threshold=lower_speed_threshold,
                higher_speed_threshold=higher_speed_threshold,
            )
        )
        complex_filters = complex_filters.join(
            self._filter_inconsistent_travel_times(dataset_in=data)
        )
        complex_filters = complex_filters.join(
            self._filter_overlapping_trips(dataset=data)
        )
        complex_filters = complex_filters.join(
            self._filter_consistent_hours(dataset=data)
        )
        complex_filters = complex_filters.join(
            self._filter_zero_length_trips(dataset=data)
        )
        return complex_filters

    @staticmethod
    def _filter_consistent_hours(dataset) -> pd.Series:
        """
        Filtering out records where starting timestamp is before end timestamp. These observations are data errors.

        Args:
            dataset : dataframe on which we would perform operations to find out the erroneous data

        Returns:
            pd.Series: Boolean Series indicating erroneous rows (trips having start time after end time) with False
        """
        ser = dataset["timestamp_start"] <= dataset["timestamp_end"]
        ser.name = "trip_start_after_end"
        return ser

    @staticmethod
    def _filter_zero_length_trips(dataset) -> pd.Series:
        """
        Filter out trips that start and end at same hour and minute but are not ending on next day (no 24-hour
        trips).

        Args:
            dataset : dataframe on which we would perform operations to find out the erroneous data

        Returns:
            pd.Series: Boolean Series indicating erroneous rows (trips having exact same start time and end time) with False
        """
        ser = ~(
            (dataset.loc[:, "trip_start_hour"] == dataset.loc[:, "trip_end_hour"])
            & (dataset.loc[:, "trip_start_minute"] == dataset.loc[:, "trip_end_minute"])
            & (~dataset.loc[:, "trip_end_next_day"])
        )
        ser.name = "is_no_zero_length_trip"
        return ser

    def _add_string_column_from_variable(self, col_name: str, var_name: str):
        """
        Replaces each occurence of a MiD/KiD variable e.g. 1,2,...,7 for
        weekdays with an explicitly mapped string e.g. 'MON', 'TUE',...,'SUN'.

        Args:
            col_name (str): Name of the column in self.trips where the explicit string info is stored
            var_name (str): Name of the venco.py internal variable given in dev_config/dataparsers['data_variables']
        """
        self.trips.loc[:, col_name] = self.trips.loc[:, var_name].replace(
            self.dev_config["dataparsers"]["replacements"][self.dataset][var_name]
        )

    @staticmethod
    def _compose_timestamp(
        data: pd.DataFrame = None,
        col_year: str = None,
        col_week: str = None,
        col_day: str = None,
        col_hour: str = None,
        col_min: str = None,
        col_name: str = None,
    ) -> pd.DatetimeIndex:
        """
        Generating pandas timestamp and storing in a new column

        Args:
            data (pd.DataFrame, optional): Defaults to None
            col_year (str, optional): Defaults to None
            col_week (str, optional): Defaults to None
            col_day (str, optional): Defaults to None
            col_hour (str, optional): Defaults to None
            col_min (str, optional): Defaults to None
            col_name (str, optional): Defaults to None

        Returns:
            pd.DatetimeIndex: modified dataframe with the new column containing the pandas timestamp values
        """
        data[col_name] = (
            pd.to_datetime(data.loc[:, col_year], format="%Y")
            + pd.to_timedelta(data.loc[:, col_week] * 7, unit="days")
            + pd.to_timedelta(data.loc[:, col_day], unit="days")
            + pd.to_timedelta(data.loc[:, col_hour], unit="hour")
            + pd.to_timedelta(data.loc[:, col_min], unit="minute")
        )
        return data

    def _compose_start_and_end_timestamps(self):
        """
        _summary_
        """
        self._compose_timestamp(
            data=self.trips,
            col_year="trip_start_year",
            col_week="trip_start_week",
            col_day="trip_start_weekday",
            col_hour="trip_start_hour",
            col_min="trip_start_minute",
            col_name="timestamp_start",
        )
        self._compose_timestamp(
            data=self.trips,
            col_year="trip_start_year",
            col_week="trip_start_week",
            col_day="trip_start_weekday",
            col_hour="trip_end_hour",
            col_min="trip_end_minute",
            col_name="timestamp_end",
        )

    @staticmethod
    def _update_end_timestamp(trips):
        """
        Updates the end timestamp for overnight trips adding one day.


        Args:
            trips : Dataframe containing details about the trip

        Returns:
            _type_: Returns the modified dataframe after adding an extra day because of the overnight journey
        """
        ends_following_day = trips["trip_end_next_day"] == 1
        trips.loc[ends_following_day, "timestamp_end"] = trips.loc[
            ends_following_day, "timestamp_end"
        ] + pd.offsets.Day(1)
        return trips

    def _assign_season(self, trips):
        """
        Function that assigns a season and recalculates electric consumption by the user choosen season, 
        if all is selected the season is assigned oriented on the timestamp_end of each trip

        Args:
            trips (_type_): _description_
        """

        if self.season == "all" and self.user_config["global"]["consider_temperature_cycle_dependency"]["annual"] is True:
            trips['season'] = trips['timestamp_end'].apply(
                self.__identify_season)
        elif pd.Series(self.season).isin(['winter', 'spring', 'fall', 'summer'])[0] and self.user_config["global"]["consider_temperature_cycle_dependency"]["annual"] is False:
            trips["season"] = self.season
        elif self.season == "None":
            trips['season'] = "not assigned"
        elif pd.Series(self.season).isin(['winter', 'spring', 'fall', 'summer'])[0] and self.user_config["global"]["consider_temperature_cycle_dependency"]["annual"] is True:
            raise ValueError('You can not generate a profile of one season with annual selection')
        else:
            raise ValueError('The season you specified does not match any known season or option')
        return self.season

    def __identify_season(self, timestamp: int):
        """
        Function returns a string with the correct season based on the input timestamp

        Args:
            timestamp (int): datetime timestamp

        Returns:
            str: season as string
        """
        day = int(pd.to_datetime(timestamp, format="%Y-%m-%d ").dayofyear)

        if (day >= winter_start) or (day < spring_start):
            return str("winter")
        elif (day >= spring_start) and (day < summer_start):
            return str("spring")
        elif (day >= summer_start) and (day < fall_start):
            return str("summer")
        else:
            return str("fall")

    def filter_for_season(self, trips):
        self.trips = trips[trips['season'] == self.user_config[
            'dataparsers']['seasons']['season_filter']]
        print(
            f"{(len(trips)-len(self.trips))} rows ({(len(self.trips)/(len(trips)-len(self.trips))*100):.2f}%) have been dropped because they did not took place in your choosed season.")
        return self.trips

    def _harmonise_variables_unique_id_names(self):
        """
        Harmonises ID variables for all datasets.
        """
        self.trips["unique_id"] = (
            self.trips[
                str(self.dev_config["dataparsers"]["id_variables_names"][self.dataset])
            ]
        ).astype(int)
        logging.info("Finished harmonisation of ID variables.")

    def _subset_vehicle_segment(self):
        """
        Dividing the vehicles into vechicle segments (eg: S, M, L in case of VF and car, van truck etc. in case of KiD).
        """
        if self.user_config["dataparsers"]["subset_vehicle_segment"]:
            self.trips = self.trips[
                self.trips["vehicle_segment_string"]
                == self.user_config["dataparsers"]["vehicle_segment"][self.dataset]
            ].reset_index(drop=True)
            segment = self.user_config["dataparsers"]["vehicle_segment"][self.dataset]
            n_vehicles = len(self.trips.unique_id.unique())
            logging.info(
                f"The subset contains only vehicles of the class {segment} for a total of {n_vehicles} individual "
                f"vehicles."
            )

    def _subset_area_type(self):
        if self.user_config["dataparsers"]["subset_area_type"]:
            self.trips = self.trips[
                self.trips["area_type"]
                == self.user_config["dataparsers"]["area_type"][self.dataset]
            ].reset_index(drop=True)
            area_type = self.user_config["dataparsers"]["area_type"][self.dataset]
            n_vehicles = len(self.trips.unique_id.unique())
            logging.info(
                f"The subset contains only vehicles for the area type {area_type} for a total of {n_vehicles} individual "
                f"vehicles."
            )



from typing import Union

from ...core.dataparsers.parseMiD import ParseMiD
from ...core.dataparsers.parseKiD import ParseKiD
from ...core.dataparsers.parseVF import ParseVF


def parse_data(configs: dict) -> Union[ParseMiD, ParseKiD, ParseVF]:
    """
    Wrapper function to process the instatiated DataParser class. The function calls the children classes which are dataset specific.

    Args:
        configs (dict): A dictionary containing a user_config dictionary and a dev_config dictionary.

    Returns:
        Union[ParseMiD, ParseKiD, ParseVF]: Children classes of the DataParser class to parse specific mobility datasets.
    """
    dataset = configs["user_config"]["global"]["dataset"]
    delegate = {"MiD08": ParseMiD, "MiD17": ParseMiD, "KiD": ParseKiD, "VF": ParseVF}
    return delegate[dataset](configs=configs, dataset=dataset)



__maintainer__ = "Niklas Wulff, Fabia Miorelli"
__license__ = "BSD-3-Clause"

import logging

from pathlib import Path

import numpy as np
import pandas as pd

from ..core.profileaggregators import ProfileAggregator
from ..utils.metadata import read_metadata_config, write_out_metadata
from ..utils.utils import create_file_name, write_out


class PostProcessor:
    def __init__(self, configs: dict, profiles: ProfileAggregator, data: dict):
        """
        This class contains functions to post-process aggregated venco.py profiles for cloning weekly profiles
        to year and normalizing it with different normalization bases.

        In the PostProcessor, two steps happen. First, the aggregated weekly timeseries for the fleet are translated
        into annual timeseries by cloning e.g. a week by around 52 times to span a full year. The second purpose is to
        normalize the aggregated fleet profiles in order to provide profiles independent of the specific input and to be
        able to scale the venco.py output in a consecutive stept with feet scenarios or annual energy demands. For the
        flow profiles, two normalization bases are applied: The drain and uncontrolled charging profiles are normalized
        to the annual energy volume of the profiles to be scaled with an annual energy demand. The charging power
        profile is normalized by the number of vehicles to retrieve a rated charging power per representative vehicle.
        The two state profiles - minimum and maximum battery level profiles - are normalized using the battery capacity
        given in the flexestimaros part of the user_config.


        Args:
            configs (dict): Dictionary holding a user_config and a dev_config dictionary
            profiles (ProfileAggregator): An instance of class ProfileAggregator
        """
        self.user_config = configs["user_config"]
        self.dev_config = configs["dev_config"]
        self.dataset = self.user_config["global"]["dataset"]
        self.upscaling_base = self.user_config["postprocessor"]["upscaling_base"]
        self.time_resolution = self.user_config["diarybuilders"]["time_resolution"]
        self.time_delta = pd.timedelta_range(
            start="00:00:00", end="24:00:00", freq=f"{self.time_resolution}T"
        )
        self.time_index = list(self.time_delta)
        self.data = data
        self.vehicle_numbers = data['vehicles_number']
        self.activities = profiles.activities
        self.seasons = ["winter", "spring", "summer", "fall"]
        self.season = self.user_config["global"]["consider_temperature_cycle_dependency"]["season"] 
        self.profile_names = (
            "drain",
            "uncontrolled_charging",
            "charging_power",
            "max_battery_level",
            "min_battery_level",
        )

        self.drain = profiles.drain_weekly
        self.charging_power = profiles.charging_power_weekly
        self.uncontrolled_charging = profiles.uncontrolled_charging_weekly
        self.max_battery_level = profiles.max_battery_level_weekly
        self.min_battery_level = profiles.min_battery_level_weekly

        self.drain_normalised = pd.DataFrame()
        self.charging_power_normalised = pd.DataFrame()
        self.uncontrolled_charging_normalised = pd.DataFrame()
        self.max_battery_level_normalised = pd.DataFrame()
        self.min_battery_level_normalised = pd.DataFrame()

        self.drain_normalised_annual = pd.DataFrame()
        self.charging_power_normalised_annual = pd.DataFrame()
        self.uncontrolled_charging_normalised_annual = pd.DataFrame()
        self.max_battery_level_normalised_annual = pd.DataFrame()
        self.min_battery_level_normalised_annual = pd.DataFrame()

        self.drain_fleet = pd.DataFrame()
        self.uncontrolled_charging_fleet = pd.DataFrame()
        self.charging_power_fleet = pd.DataFrame()
        self.max_battery_level_fleet = pd.DataFrame()
        self.min_battery_level_fleet = pd.DataFrame()

        self.annual_profiles = {}
        self.normalised_profiles = {}
        self.fleet_profiles = {}

    def __week_to_annual_profile(self, profile: pd.Series) -> pd.Series:
        """
        This function clones a profile with a given temporal resolution from a weekly profile to the full year.
        Overreaching time intervals are being truncated at the year border (December, 31st). Currently, gap years are
        not considered.

        Args:
            profile (pd.Series): Any profile spanning one week in variable resolution

        Returns:
            pd.Series: Annual profile in the same temporal resolution as the input profile
        """
        start_weekday = self.user_config["postprocessor"]["start_weekday"]  # (1: Monday, 7: Sunday)
        n_timeslots_per_day = len(list(self.time_index))
        if self.user_config["global"][
            "consider_temperature_cycle_dependency"]["annual"] and self.user_config["global"]["consider_temperature_cycle_dependency"]["season"] != 'None':
            annual = profile.iloc[((start_weekday - 1) * (n_timeslots_per_day - 1)):]
            seasonal = pd.DataFrame()
            for season in self.seasons:
                annual_season = annual[season]
                annual_season = pd.concat([annual_season] * 13, ignore_index = True)
                seasonal = pd.concat([seasonal, annual_season], ignore_index=True)
            # add a few bins of winter to match a whole year
            missing_bins = (n_timeslots_per_day - 1) * 365 - len(seasonal)
            annual_season = annual['winter'].head(missing_bins)
            seasonal = pd.concat([seasonal, annual_season], ignore_index = True)
            annual = pd.DataFrame(seasonal)

        else:
            # Shift input profiles to the right weekday and start with first bin of chosen weekday
            annual = profile.iloc[((start_weekday - 1) * (n_timeslots_per_day - 1)) :]
            annual = pd.DataFrame(annual.iloc[:, 0].to_list() * 53)
            annual.drop(annual.tail(len(annual) - (n_timeslots_per_day - 1) * 365).index)
        return annual

    @staticmethod
    def __normalise_flows(profile: pd.Series) -> pd.Series:
        """
        Function to normalise a timeseries according to its annual sum. Used in venco.py for normalisation of
        uncontrolled charging and drain profiles.

        Args:
            profile (pd.Series): Weekly profile to be normalised

        Returns:
            pd.Series: Normalized timeseries
        """
        return profile / (profile.sum() / 7 * 365)

    @staticmethod
    def __normalise_states(profile: pd.Series, base: int) -> pd.Series:
        """
        Function to normalise a state profile according to a baseline value. Used in venco.py for normalisation of the
        minimum and maximum battery level profiles based on the vehicle battery capacity assumed in the flexestimator
        config.

        Args:
            profile (pd.Series): State profile to be normalised
            base (int): normalisation basis, e.g. battery capacity in kWh

        Returns:
            pd.Series: Normalized battery level between 0 and 1
        """
        return profile / base

    @staticmethod
    def __normalise_based_on_vehicle_numbers(
        profile: pd.Series, base: int, time_delta
    ) -> pd.Series:
        """
        Function to normalise a timeseries according to a baseline value for each weekday. Used in venco.py for
        normalisation of profiles based on the number of vehicles for each weekday.

        Args:
            profile (pd.Series): Profile to be normalised
            base (int): normalisation basis, e.g. number of vehicles
            time_delta (pd.timedelta_range): Temporal resolution of the run

        Returns:
            pd.Series: Charging power profile of the fleet normalized to the number of vehicles of the fleet
        """
        profile_normalised = []
        for day in range(1, 8):
            start = day * len(time_delta)
            end = start + len(time_delta)
            profile_day = profile[start:end] / (base[day])
            if profile_normalised == []:
                profile_normalised = profile_day.to_list()
            else:
                profile_normalised.extend(profile_day.to_list())
        profile = pd.Series(profile_normalised)
        return profile

    def __write_output(self, profile_name: str, profile: pd.Series, filename_id: str):
        """
        Write out the profile given in profile adding the profile_name and a filename_id (overarching file label) to the
        written file.

        Args:
            profile_name (str): A string describing the name of a profile to write
            profile (pd.Series): A profile to write to disk
            filename_id (str): An additional overarching label provided to the file written to disk
        """
        root = Path(self.user_config["global"]["absolute_path"]["vencopy_root"])
        folder = self.dev_config["global"]["relative_path"]["processor_output"]
        self.dev_config["global"]["additional_label"] = profile_name
        file_name = create_file_name(
            dev_config=self.dev_config,
            user_config=self.user_config,
            file_name_id=filename_id,
            dataset=self.dataset,
        )
        write_out(data=profile, path=root / folder / file_name)

    def create_fleet_data_structure(self):
        nuts_resolution = self.user_config["postprocessor"]["spatial"]["nuts_level"]
        spatial_resolution = self.user_config["postprocessor"]["spatial"]["country"]
        temporal_resolution = self.user_config["postprocessor"]["temporal"]
        time_resolution = self.user_config["diarybuilders"]["time_resolution"]
        eurostat_pop_file = (
            Path(self.user_config["global"]["absolute_path"]["vencopy_root"])
            / Path(self.dev_config["global"]["relative_path"]["eurostat_population"])
            / self.dev_config["global"]["files"]["eurostat_population"]
        )
        eurostat_nuts3_population = self._read_in_eurostat_data_nodes(
            eurostat_pop_file=eurostat_pop_file
        )
        data_nodes = self._keep_population_model_scope(
            eurostat_nuts3_population=eurostat_nuts3_population,
            spatial_resolution=spatial_resolution,
            temporal_resolution=temporal_resolution,
        )
        population_shares = self._calculate_population_share(
            data_nodes=data_nodes,
            nuts_resolution=nuts_resolution,
            temporal_resolution=temporal_resolution,
        )
        fleet_data = self._adjust_fleet_data_structure(
            time_resolution=time_resolution, population_shares=population_shares
        )
        return fleet_data

    @staticmethod
    def _scale_up_fleet_size(
        fleet_size: dict,
        fleet_data_structure: pd.DataFrame,
        normalised_profile: pd.DataFrame,
    ) -> pd.DataFrame:
        """
        _summary_

        Args:
            fleet_size (dict): _description_
            fleet_data_structure (pd.DataFrame): _description_
            normalised_profile (pd.DataFrame): _description_

        Returns:
            pd.DataFrame: _description_
        """
        normalised_profile.index.name = "timestep"
        fleet_data = fleet_data_structure.groupby(level=[0], group_keys=False).apply(
            lambda x: x.mul(normalised_profile[0], axis=0)
        )
        fleet_size = pd.DataFrame.from_dict(fleet_size)
        for year in fleet_size.columns:
            fleet_data = fleet_data.groupby(level=[0], group_keys=False).apply(
                lambda x: x * fleet_size[year].values[0]
            )
        return fleet_data

    @staticmethod
    def _scale_up_annual_demand(
        annual_demand: dict,
        fleet_data_structure: pd.DataFrame,
        normalised_profile: pd.DataFrame,
    ) -> pd.DataFrame:
        """
        _summary_

        Args:
            fleet_size (dict): _description_
            fleet_data_structure (pd.DataFrame): _description_
            normalised_profile (pd.DataFrame): _description_

        Returns:
            pd.DataFrame: _description_
        """
        normalised_profile.index.name = "timestep"
        fleet_data = fleet_data_structure.groupby(level=[0], group_keys=False).apply(
            lambda x: x.mul(normalised_profile[0], axis=0)
        )
        annual_demand = pd.DataFrame.from_dict(annual_demand)
        for year in annual_demand.columns:
            fleet_data = fleet_data.groupby(level=[0], group_keys=False).apply(
                lambda x: x * annual_demand[year].values[0]
            )
        return fleet_data

    @staticmethod
    def _read_in_eurostat_data_nodes(eurostat_pop_file: Path) -> pd.DataFrame:
        """
        Read-in eurostat population file. Updated file can be downloaded at:
        https://ec.europa.eu/eurostat/web/population-demography/population-projections/database

        Args:
            eurostat_pop_file (Path): path to the venco.py input folder containing the eurostat population

        Returns:
            pd.DataFrame: _description_
        """
        eurostat_nuts3_population = pd.read_csv(eurostat_pop_file, sep="\t")
        eurostat_nuts3_population = eurostat_nuts3_population.rename(
            columns=lambda x: x.strip()
        )
        headers = eurostat_nuts3_population[
            "projection,age,sex,unit,geo\\time"
        ].str.split(",", expand=True)
        headers = headers.rename(
            columns={0: "projection", 1: "age", 2: "sex", 3: "unit", 4: "nuts_3"}
        )
        eurostat_nuts3_population = pd.concat(
            [headers, eurostat_nuts3_population], axis=1
        )
        eurostat_nuts3_population.drop(
            "projection,age,sex,unit,geo\\time", axis=1, inplace=True
        )
        eurostat_nuts3_population = eurostat_nuts3_population.loc[
            (eurostat_nuts3_population["age"] == "TOTAL")
            & (eurostat_nuts3_population["sex"] == "T")
            & (eurostat_nuts3_population["projection"] == "NIRMIGR")
        ]
        eurostat_nuts3_population.drop(
            ["projection", "age", "sex", "unit"], axis=1, inplace=True
        )
        return eurostat_nuts3_population.reset_index(drop=True)

    @staticmethod
    def _keep_population_model_scope(
        eurostat_nuts3_population: pd.DataFrame,
        spatial_resolution: str,
        temporal_resolution: list,
    ) -> pd.DataFrame:
        """
        Subsets the eurostat dataframe containing the population for the required spatial and temporal resolution.

        Args:
            eurostat_nuts3_population (pd.DataFrame): _description_
            spatial_resolution (str): _description_
            temporal_resolution (list): _description_

        Returns:
            pd.DataFrame: _description_
        """
        data_nodes = eurostat_nuts3_population[["nuts_3"] + temporal_resolution]
        data_nodes = data_nodes[data_nodes["nuts_3"].str.contains(spatial_resolution)]
        return data_nodes.reset_index(drop=True)

    @staticmethod
    def _calculate_population_share(
        data_nodes: pd.DataFrame, nuts_resolution: str, temporal_resolution: list
    ) -> pd.DataFrame:
        """
        Calculate the population share based on the given NUTS and temporal resolution.

        Args:
            data_nodes (pd.DataFrame): dataframe containing columns with absolute population number for the specified country and NUTS level
            nuts_resolution (str): NUTS level of interest
            temporal_resolution (list): list of modelling years to be considered

        Returns:
            pd.DataFrame: dataframe with share of population for the specified modelling years and NUTS level
        """
        total_population = data_nodes[
            temporal_resolution
        ].sum()  # note: option of multiple countries not possible -> another model run
        if nuts_resolution == "nuts_0":
            data_nodes["nuts_0"] = data_nodes["nuts_3"].str.slice(0, 2)
            data_nodes.drop("nuts_3", axis=1, inplace=True)
            data_nodes = data_nodes.groupby(by="nuts_0").sum()
            population_shares = data_nodes / total_population * 100
        elif nuts_resolution == "nuts_1":
            data_nodes["nuts_1"] = data_nodes["nuts_3"].str.slice(0, 3)
            data_nodes.drop("nuts_3", axis=1, inplace=True)
            data_nodes = data_nodes.groupby(by="nuts_1").sum()
            population_shares = data_nodes / total_population * 100
        elif nuts_resolution == "nuts_2":
            data_nodes["nuts_2"] = data_nodes["nuts_3"].str.slice(0, 4)
            data_nodes.drop("nuts_3", axis=1, inplace=True)
            data_nodes = data_nodes.groupby(by="nuts_2").sum()
            population_shares = data_nodes / total_population * 100
        elif nuts_resolution == "nuts_3":
            data_nodes = data_nodes.groupby(by="nuts_3").sum()
            population_shares = data_nodes / total_population * 100
        return population_shares

    @staticmethod
    def _adjust_fleet_data_structure(time_resolution, population_shares):
        """
        Creates dataframe structure in which the profiles are stored for the predefined NUTS resolution and for the specified modelling years.

        Args:
            time_resolution (int): venco.py time resolution for timeseries
            population_shares (pd.DataFrame): Eurostat population share for the predefined temporal and geographical resolution
        """
        fleet = population_shares.transpose()
        fleet.index.name = "year"
        fleet.index = fleet.index.astype(int)
        fleet = fleet.reset_index()
        fleet = pd.DataFrame(
            np.repeat(fleet.values, int(24 * 365 * 60 / time_resolution), axis=0),
            columns=fleet.columns,
        )
        df_years_list = []
        for year in fleet["year"].unique():
            df_year = fleet[fleet["year"] == int(year)].copy()
            df_year.reset_index(inplace=True)
            df_year.drop(["index"], axis=1, inplace=True)
            df_year.index.name = "timestep"
            multi_idx = pd.MultiIndex.from_arrays([df_year["year"], df_year.index])
            df_year.index = multi_idx
            df_year.drop(["year"], axis=1, inplace=True)
            df_years_list.append(df_year)
        fleet = pd.concat(df_years_list)
        return fleet

    def generate_metadata(self, metadata_config: dict, file_name: str) -> dict:
        """
        _summary_

        Args:
            metadata_config (dict): _description_
            file_name (str): _description_

        Returns:
            dict: _description_
        """
        metadata_config["name"] = file_name
        if "normalised" in file_name:
            metadata_config["title"] = (
                "Annual normalised timeseries with venco.py output profiles"
            )
            metadata_config["description"] = "Annual normalised timeseries."
        elif "annual" in file_name:
            metadata_config["title"] = (
                "Annual timeseries with venco.py output profiles at fleet level"
            )
            metadata_config["description"] = "Annual timeseries at fleet level."
        metadata_config["sources"] = [
            f for f in metadata_config["sources"] if f["title"] in self.dataset
        ]
        reference_resource = metadata_config["resources"][0]
        this_resource = reference_resource.copy()
        this_resource["name"] = file_name.rstrip(".metadata.yaml")
        this_resource["path"] = file_name
        if "normalised" in file_name:
            these_fields = [
                f
                for f in reference_resource["schema"][self.dataset]["fields"][
                    "postprocessors"
                ]["normalised"]
            ]
        elif "annual" in file_name:
            these_fields = [
                f
                for f in reference_resource["schema"][self.dataset]["fields"][
                    "postprocessors"
                ]["annual"]
            ]
        this_resource["schema"] = {"fields": these_fields}
        metadata_config["resources"].pop()
        metadata_config["resources"].append(this_resource)
        return metadata_config

    def _write_metadata(self, file_name: str):
        """
        _summary_

        Args:
            file_name (str): _description_
        """
        metadata_config = read_metadata_config()
        class_metadata = self.generate_metadata(
            metadata_config=metadata_config, file_name=file_name.name
        )
        write_out_metadata(metadata_yaml=class_metadata, file_name=file_name)

    def create_annual_profiles(self):
        """
        Wrapper function to clone the five main venco.py profiles from weekly profiles to annual
        profiles and write them to disk. This function is meant to be called in the run.py.
        """
        if self.user_config["profileaggregators"]["aggregation_timespan"] == "daily":
            logging.info(
                "The annual profiles cannot be generated as the aggregation was performed over a single day."
            )
        elif (self.user_config["global"]["consider_temperature_cycle_dependency"]["season"] in self.seasons) & self.user_config["global"]["consider_temperature_cycle_dependency"]["annual"]:
            logging.info(
                "The annual profiles cannot be generated as only a specific season was selected."
            )
        else:

            self.drain_normalised_annual["drain"] = self.__week_to_annual_profile(profile=self.drain_normalised)
            self.uncontrolled_charging_normalised_annual["uncontrolled_charging"] = self.__week_to_annual_profile(profile=self.uncontrolled_charging_normalised)
            self.charging_power_normalised_annual["charging_power"] = self.__week_to_annual_profile(profile=self.charging_power_normalised)
            self.max_battery_level_normalised_annual["max_battery_level"] = self.__week_to_annual_profile(profile=self.max_battery_level_normalised)
            self.min_battery_level_normalised_annual["min_battery_level"] = self.__week_to_annual_profile(profile=self.min_battery_level_normalised)
        self.annual_profiles = (
            self.drain_normalised_annual,
            self.uncontrolled_charging_normalised_annual,
            self.charging_power_normalised_annual,
            self.max_battery_level_normalised_annual,
            self.min_battery_level_normalised_annual,
        )
        profile_names = self.profile_names
        if self.user_config["global"]["write_output_to_disk"]["processor_output"][
            "normalised_annual_profiles"
        ]:
            for profile_name, profile in zip(profile_names, self.annual_profiles):
                self.__write_output(
                    profile_name=profile_name,
                    profile=profile,
                    filename_id="output_postprocessor_annual",
                )
        root = Path(self.user_config["global"]["absolute_path"]["vencopy_root"])
        folder = self.dev_config["global"]["relative_path"]["processor_output"]
        file_name = (
            "vencopy_output_postprocessor_annual_normalised_"
            + self.dataset
            + ".metadata.yaml"
        )
        self._write_metadata(file_name=root / folder / file_name)

    def normalise_profiles(self):
        """
        Wrapper function to normalise the five venco.py profiles to default normalisation bases. These are
        the year-sum of the profiles for drain and uncontrolled charging, the vehicle fleet for charging power and the
        battery capacity for min and max battery level profiles.
        This profile is supposed to be called in the venco.py workflow after PostProcessor instantiation.
        """
        if self.user_config["profileaggregators"]["aggregation_timespan"] == "daily":
            raise ValueError(
                "The annual profiles cannot be normalised as the aggregation was performed over a single day."
            )
        else:
            if self.user_config["global"]["consider_temperature_cycle_dependency"]["annual"] & (self.season != 'None'):
                for season in self.seasons:
                    self.drain_normalised[season] = self.__normalise_based_on_vehicle_numbers(profile=self.drain[season], base=self.vehicle_numbers[season]["total_amount_vehicle"], time_delta=self.time_delta)
                    self.uncontrolled_charging_normalised[season] = self.__normalise_based_on_vehicle_numbers(profile=self.uncontrolled_charging[season], base=self.vehicle_numbers[season]["total_amount_vehicle"], time_delta=self.time_delta)
                    self.charging_power_normalised[season] = self.__normalise_based_on_vehicle_numbers(profile=self.charging_power[season], base=self.vehicle_numbers[season]["total_amount_vehicle"], time_delta=self.time_delta)
                    self.max_battery_level_normalised[season] = self.__normalise_states(profile=self.max_battery_level[season], base=1)
                    self.min_battery_level_normalised[season] = self.__normalise_states(profile=self.min_battery_level[season], base=1)
            elif self.upscaling_base == "fleet_size":
                self.drain_normalised["drain"] = self.__normalise_based_on_vehicle_numbers(profile=self.drain["drain"], base=self.vehicle_numbers["total_amount_vehicle"], time_delta=self.time_delta)
                self.uncontrolled_charging_normalised["uncontrolled_charging"] = self.__normalise_based_on_vehicle_numbers(profile=self.uncontrolled_charging["uncontrolled_charging"], base=self.vehicle_numbers["total_amount_vehicle"], time_delta=self.time_delta)
                self.charging_power_normalised["charging_power"] = self.__normalise_based_on_vehicle_numbers(profile=self.charging_power["charging_power"], base=self.vehicle_numbers["total_amount_vehicle"], time_delta=self.time_delta)
                self.max_battery_level_normalised["max_battery_level"] = self.__normalise_states(profile=self.max_battery_level["max_battery_level"], base=1)
                self.min_battery_level_normalised["min_battery_level"] = self.__normalise_states(profile=self.min_battery_level["min_battery_level"], base=1)
            elif self.upscaling_base == "annual_demand":
                self.drain_normalised["drain"] = self.__normalise_flows(profile=self.drain["drain"])
                self.uncontrolled_charging_normalised["uncontrolled_charging"] = self.__normalise_flows(profile=self.uncontrolled_charging["uncontrolled_charging"])
                self.charging_power_normalised["charging_power"] = self.__normalise_based_on_vehicle_numbers(profile=self.charging_power["charging_power"], base=self.vehicle_numbers["total_amount_vehicle"], time_delta=self.time_delta)
                self.max_battery_level_normalised["max_battery_level"] = self.__normalise_states(profile=self.max_battery_level["max_battery_level"], base=1)
                self.min_battery_level_normalised["min_battery_level"] = self.__normalise_states(profile=self.min_battery_level["min_battery_level"], base=1)
            else:
                NotImplementedError(
                    "The normalisation can either be based on the number of vehicles in the fleet or on the annual demand."
                )
            self.normalised_profiles = (
                self.drain_normalised,
                self.uncontrolled_charging_normalised,
                self.charging_power_normalised,
                self.max_battery_level_normalised,
                self.min_battery_level_normalised,
            )
            profile_names = self.profile_names
            if self.user_config["global"]["write_output_to_disk"]["processor_output"][
                "normalised_weekly_profiles"
            ]:
                for profile_name, profile in zip(
                    profile_names, self.normalised_profiles
                ):
                    self.__write_output(
                        profile_name=profile_name,
                        profile=profile,
                        filename_id="output_postprocessor_normalised",
                    )
            root = Path(self.user_config["global"]["absolute_path"]["vencopy_root"])
            folder = self.dev_config["global"]["relative_path"]["processor_output"]
            file_name = (
                "vencopy_output_postprocessor_weekly_normalised_"
                + self.dataset
                + ".metadata.yaml"
            )
            self._write_metadata(file_name=root / folder / file_name)

    def generate_fleet_profiles(self):
        """
        Wrapper function to scale up the five main venco.py profiles from weekly profiles to fleet
        level and write them to disk. This function is meant to be called in the run.py.
        """
        if self.user_config["profileaggregators"]["aggregation_timespan"] == "daily" or self.user_config["global"]["consider_temperature_cycle_dependency"]["season"] != "all":
            logging.info(
                "The fleet profiles cannot be generated as the aggregation was performed over a single day or only one season was selected."
            )
        else:
            fleet_data_structure = self.create_fleet_data_structure()
            if self.upscaling_base == "fleet_size":
                fleet_size = self.user_config["postprocessor"]["fleet_size_nuts_0"]
                self.drain_fleet = self._scale_up_fleet_size(normalised_profile=self.drain_normalised_annual["drain"], fleet_size=fleet_size, fleet_data_structure=fleet_data_structure)
                self.uncontrolled_charging_fleet = self._scale_up_fleet_size(normalised_profile=self.uncontrolled_charging_normalised_annual["uncontrolled_charging"], fleet_size=fleet_size, fleet_data_structure=fleet_data_structure)
                self.charging_power_fleet = self._scale_up_fleet_size(normalised_profile=self.charging_power_normalised_annual["charging_power"], fleet_size=fleet_size, fleet_data_structure=fleet_data_structure)
                self.max_battery_level_fleet = self._scale_up_fleet_size(normalised_profile=self.max_battery_level_normalised_annual["max_battery_level"], fleet_size=fleet_size, fleet_data_structure=fleet_data_structure)
                self.min_battery_level_fleet = self._scale_up_fleet_size(normalised_profile=self.min_battery_level_normalised_annual["min_battery_level"], fleet_size=fleet_size, fleet_data_structure=fleet_data_structure)
            elif self.upscaling_base == "annual_demand":
                annual_demand = self.user_config["postprocessor"]["annual_demand_nuts_0"][0] # in TWh
                vehicle_specific_consumption = self.user_config["flexestimators"]["electric_consumption"]["general"] / 100 / 1e9  # consumption from kWh/100 km to kWh/km to TWh/km
                vehicle_yearly_mileage = self.user_config["postprocessor"]["yearly_mileage"]  # km/year/vehicle
                fleet_size = annual_demand / vehicle_specific_consumption / vehicle_yearly_mileage
                self.drain_fleet = self._scale_up_annual_demand(normalised_profile=self.drain_normalised_annual["drain"], annual_demand=annual_demand, fleet_data_structure=fleet_data_structure)
                self.uncontrolled_charging_fleet = self._scale_up_annual_demand(normalised_profile=self.uncontrolled_charging_normalised_annual["uncontrolled_charging"], annual_demand=annual_demand, fleet_data_structure=fleet_data_structure)
                self.charging_power_fleet = self._scale_up_fleet_size(normalised_profile=self.charging_power_normalised["charging_power"], fleet_size=fleet_size, fleet_data_structure=fleet_data_structure)
                self.max_battery_level_fleet = self._scale_up_fleet_size(normalised_profile=self.max_battery_level_normalised_annual["max_battery_level"], fleet_size=fleet_size, fleet_data_structure=fleet_data_structure)
                self.min_battery_level_fleet = self._scale_up_fleet_size(normalised_profile=self.min_battery_level_normalised_annual["min_battery_level"], fleet_size=fleet_size, fleet_data_structure=fleet_data_structure)
            self.fleet_profiles = (
                self.drain_fleet,
                self.uncontrolled_charging_fleet,
                self.charging_power_fleet,
                self.max_battery_level_fleet,
                self.min_battery_level_fleet,
            )
            profile_names = self.profile_names
            if self.user_config["global"]["write_output_to_disk"]["processor_output"][
                "fleet_annual_profiles"
            ]:
                for profile_name, profile in zip(profile_names, self.fleet_profiles):
                    self.__write_output(
                        profile_name=profile_name,
                        profile=profile,
                        filename_id="output_postprocessor_fleet",
                    )
            root = Path(self.user_config["global"]["absolute_path"]["vencopy_root"])
            folder = self.dev_config["global"]["relative_path"]["processor_output"]
            file_name = (
                "vencopy_output_postprocessor_annual_fleet_"
                + self.dataset
                + ".metadata.yaml"
            )
            self._write_metadata(file_name=root / folder / file_name)



__maintainer__ = "Fabia Miorelli"
__license__ = "BSD-3-Clause"

import pytest
import pandas as pd
import numpy as np


from ....vencopy.core.dataparsers.parkinference import ParkInference
from ....vencopy.core.dataparsers.parseKiD import ParseKiD

# NOT TESTED: process(), __add_string_columns(), _load_unencrypted_data()

@pytest.fixture
def parse_kid_instance():
    configs = {
        'user_config': {
            'global': {
                'dataset': 'dataset1',
                'debug': False,
                'number_lines_debug': 5000,
                'absolute_path': {
                    'dataset1': '/path/to/dataset1',
                    'dataset2': '/path/to/dataset2'
                    },
                'consider_temperature_cycle_dependency': {
                    'annual': True,
                    'season': 'all',
                    'daily': False
                }
                },
        "dataparsers": {
            "subset_vehicle_segment": True,
            "vehicle_segment": {
                "dataset1": "Car",
                "dataset2": "L"
            },
            'seasons': {
                'filter_season_post_assignment': False,
                'season_filter': "summer"
                },
                }
            },
        'dev_config': {
            'global': {
                    'files': {
                        "dataset1": {
                            "trips_data_raw": "trips01.csv"
                                    },
                        "dataset2": {
                            "trips_data_raw": "trips02.csv"
                                    }
                            }
                        },
            'dataparsers': {
                'data_variables': {
                    'dataset': ['dataset1', 'dataset2', 'dataset3'],
                    'var1': ['var1dataset1', 'var1dataset2', 'var1dataset3'],
                    'var2': ['var2dataset1', 'var2dataset2', 'var2dataset3'],
                    },
                "filters": {
                        "dataset1": {
                            "filter1": [1, 2, 3],
                            "filter2": ["A", "B", "C"]
                                    },
                        "dataset2": {
                            "filter3": [True, False, True]
                                    }
                                },
                "input_data_types": {
                    "dataset1": {
                        "var1dataset1": int,
                        "var2dataset1": str
                    }},
                "replacements":{
                    "dataset1": {
                        "var1dataset1": {
                            "1": "One",
                            "2": "Two",
                            "3": "Three"
                        }}},
                "id_variables_names": {
                    "dataset1": "id01",
                    "dataset2": "id02"
                            }
                        }
                    }
        }
    dataset = "dataset1"
    return ParseKiD(configs, dataset)


def test_parse_kid_init(parse_kid_instance):
    assert isinstance(parse_kid_instance, ParseKiD)
    assert isinstance(parse_kid_instance.park_inference, ParkInference)


def test_change_separator():
    trips = pd.DataFrame({'trip_distance': ['1234,56', '789,01'],
                          'trip_weight': ['3.45', '6,789']})
    result_df = ParseKiD._change_separator(trips=trips)

    expected_result = pd.DataFrame({'trip_distance': ['1234.56', '789.01'],
                                    'trip_weight': ['3.45', '6.789']})
    pd.testing.assert_frame_equal(result_df, expected_result)


def test_extract_timestamps():
    trips = pd.DataFrame({'trip_start_date': ['01.09.2023', '02.09.2023', '03.09.2023'],
                          'trip_start_clock': ['08:00', '09:30', '10:15'],
                          'trip_end_clock': ['08:45', '10:00', '11:30']})

    result = ParseKiD._extract_timestamps(trips=trips)

    expected_result = pd.DataFrame({
        'trip_start_date': pd.to_datetime(['2023-09-01', '2023-09-02', '2023-09-03']),
        'trip_start_clock': ['08:00', '09:30', '10:15'],
        'trip_end_clock': ['08:45', '10:00', '11:30'],
        'trip_start_year': [2023, 2023, 2023],
        'trip_start_month': [9, 9, 9],
        'trip_start_day': [1, 2, 3],
        'trip_start_weekday': [4, 5, 6],
        'trip_start_week': [35, 35, 35],
        'trip_start_hour': [8, 9, 10],
        'trip_start_minute': [0, 30, 15],
        'trip_end_hour': [8, 10, 11],
        'trip_end_minute': [45, 0, 30],
    })
    expected_result['trip_start_week'] = pd.Series(expected_result["trip_start_week"], dtype="int")
    pd.testing.assert_frame_equal(result, expected_result, check_dtype=False)


def test_update_end_timestamp():
    trips = pd.DataFrame({
        "timestamp_start": pd.DatetimeIndex(["2023-09-01 08:00:00", "2023-09-01 09:30:00", "2023-09-02 23:15:00"]),
        "timestamp_end": pd.DatetimeIndex(["2023-09-01 08:45:00", "2023-09-01 10:00:00", "2023-09-02 00:30:00"]),
    })

    result = ParseKiD._update_end_timestamp(trips)

    expected_result = pd.DataFrame({
        "timestamp_start": pd.DatetimeIndex(["2023-09-01 08:00:00", "2023-09-01 09:30:00", "2023-09-02 23:15:00"]),
        "timestamp_end": pd.DatetimeIndex(["2023-09-01 08:45:00", "2023-09-01 10:00:00", "2023-09-03 00:30:00"]),
        "trip_end_next_day": [False, False, True]})
    pd.testing.assert_frame_equal(result, expected_result)


def test_exclude_hours():
    trips = pd.DataFrame({
        'trip_start_clock': ['08:00', '-1:-1', '10:15', '-1:-1'],
        'trip_end_clock': ['08:45', '09:30', '-1:-1', '-1:-1']})

    result = ParseKiD._exclude_hours(trips=trips)

    expected_result = pd.DataFrame({
        'trip_start_clock': ['08:00'],
        'trip_end_clock': ['08:45'],
    })

    pd.testing.assert_frame_equal(result, expected_result)



__maintainer__ = "Fabia Miorelli"
__license__ = "BSD-3-Clause"

from pathlib import Path

import pandas as pd
import pytest

from ....vencopy.core.dataparsers.dataparsers import IntermediateParsing

# NOT TESTED: _complex_filters(), _compose_start_and_end_timestamps(),


@pytest.fixture
def sample_configs():
    configs = {
        'user_config': {
            'global': {
                'debug': False,
                'number_lines_debug': 5000,
                'absolute_path': {
                    'dataset1': '/path/to/dataset1',
                    'dataset2': '/path/to/dataset2'
                    },
                'consider_temperature_cycle_dependency': {
                    'annual': True,
                    'season': 'all',
                    'daily': False
                }    
                },
            'dataparsers': {
                'seasons': {
                    'filter_season_post_assignment': False,
                    'season_filter': "summer"
                    },
                'subset_vehicle_segment': True,
                'vehicle_segment': {
                    'dataset1': 'Car',
                    'dataset2': 'Car'
                    }
            },
        },
        'dev_config': {
            'global': {
                    'files': {
                        "dataset1": {
                            "trips_data_raw": "trips01.csv"
                                    },
                        "dataset2": {
                            "trips_data_raw": "trips02.csv"
                                    }
                            }
                        },
            'dataparsers': {
                'data_variables': {
                    'dataset': ['dataset1', 'dataset2', 'dataset3'],
                    'var1': ['var1dataset1', 'var1dataset2', 'var1dataset3'],
                    'var2': ['var2dataset1', 'var2dataset2', 'var2dataset3'],
                    },
                "filters": {
                        "dataset1": {
                            "filter1": [1, 2, 3],
                            "filter2": ["A", "B", "C"]
                                    },
                        "dataset2": {
                            "filter3": [True, False, True]
                                    }
                                }
                            }
                        }
                    }
    return configs


def test_intermediate_parsing_init(sample_configs):
    dataset = "dataset1"
    parser = IntermediateParsing(sample_configs, dataset)

    assert parser.user_config == sample_configs["user_config"]
    assert parser.dev_config == sample_configs["dev_config"]
    assert parser.debug == False
    assert parser.dataset == dataset
    assert parser.raw_data_path == Path("/path/to/dataset1/trips01.csv")
    assert parser.raw_data is None
    assert parser.trips is None
    assert parser.activities is None
    assert parser.filters == {'filter1': [1, 2, 3], 'filter2': ['A', 'B', 'C']}
    assert parser.filters == sample_configs["dev_config"]["dataparsers"]["filters"][dataset]
    assert parser.var_datatype_dict == {}
    assert parser.columns == ['var1dataset1', 'var2dataset1']


def test_compile_variable_list(sample_configs):
    dataset = "dataset1"
    parser = IntermediateParsing(sample_configs, dataset)

    variables = parser._compile_variable_list()
    expected_variables = ['var1dataset1', 'var2dataset1']
    assert variables == expected_variables


def test_remove_na():
    variables = ["var1", "var2", "NA"]
    IntermediateParsing._remove_na(variables)

    assert "NA" not in variables


@pytest.fixture
def intermediate_parser_instance():
    configs = {
        'user_config': {
            'global': {
                'debug': False,
                'number_lines_debug': 5000,
                'absolute_path': {
                    'dataset1': '/path/to/dataset1',
                    'dataset2': '/path/to/dataset2'
                    },
                'consider_temperature_cycle_dependency': {
                    'annual': True,
                    'season': 'all',
                    'daily': False
                }
                },
            'dataparsers': {
                'seasons': {
                    'filter_season_post_assignment': False,
                    'season_filter': "summer"
                    },
                'subset_vehicle_segment': True,
                'vehicle_segment': {
                    'dataset1': 'Car',
                    'dataset2': 'Car' 
                    }
                }
            },
        'dev_config': {
            'global': {
                    'files': {
                        "dataset1": {
                            "trips_data_raw": "trips01.csv"
                                    },
                        "dataset2": {
                            "trips_data_raw": "trips02.csv"
                                    }
                            }
                        },
            'dataparsers': {
                'data_variables': {
                    'dataset': ['dataset1', 'dataset2', 'dataset3'],
                    'var1': ['var1dataset1', 'var1dataset2', 'var1dataset3'],
                    'var2': ['var2dataset1', 'var2dataset2', 'var2dataset3'],
                    },
                "filters": {
                        "dataset1": {
                            "filter1": [1, 2, 3],
                            "filter2": ["A", "B", "C"]
                                    },
                        "dataset2": {
                            "filter3": [True, False, True]
                                    }
                                },
                "input_data_types": {
                    "dataset1": {
                        "var1dataset1": int,
                        "var2dataset1": str
                    }},
                "replacements":{
                    "dataset1": {
                        "var1dataset1": {
                            "1": "One",
                            "2": "Two",
                            "3": "Three"
                        }}},
                "id_variables_names": {
                    "dataset1": "id01",
                    "dataset2": "id02"
                            }
                        }
                    }
        }
    dataset = "dataset1"
    return IntermediateParsing(configs, dataset)


def test_select_columns(intermediate_parser_instance):
    raw_data = pd.DataFrame({
        "var1dataset1": [1.0, 2.0, 3.0],
        "var2dataset1": ["A", "B", "C"],
    })

    intermediate_parser_instance.raw_data = raw_data
    intermediate_parser_instance._select_columns()
    expected_columns = ["var1dataset1", "var2dataset1"]
    assert list(intermediate_parser_instance.trips.columns) == expected_columns


def test_convert_types(intermediate_parser_instance):
    trips_data = pd.DataFrame({
        "var1dataset1": [1.0, 2.0, 3.0],
        "var2dataset1": ["A", "B", "C"],
    })

    intermediate_parser_instance.trips = trips_data
    intermediate_parser_instance._convert_types()

    assert intermediate_parser_instance.trips["var1dataset1"].dtype == int
    assert intermediate_parser_instance.trips["var2dataset1"].dtype == object #TODO: should be string here


def test_filter_consistent_hours():
    dataset = pd.DataFrame({
        "timestamp_start": [pd.to_datetime("2023-09-12 08:00:00"), pd.to_datetime("2023-09-12 09:00:00"), pd.to_datetime("2023-09-12 09:50:00")],
        "timestamp_end": [pd.to_datetime("2023-09-12 09:00:00"), pd.to_datetime("2023-09-12 10:00:00"), pd.to_datetime("2023-09-12 09:30:00")]
    })

    result = IntermediateParsing._filter_consistent_hours(dataset)
    
    expected_result = pd.Series([True, True, False], name="trip_start_after_end")
    pd.testing.assert_series_equal(result, expected_result)


def test_filter_zero_length_trips():
    dataset = pd.DataFrame({
        "trip_start_hour": [8, 9, 10],
        "trip_start_minute": [0, 0, 0],
        "trip_end_hour": [8, 9, 10],
        "trip_end_minute": [0, 0, 1],
        "trip_end_next_day": [False, False, False]
    })

    result = IntermediateParsing._filter_zero_length_trips(dataset)
    
    expected_result = pd.Series([False, False, True], name="is_no_zero_length_trip")
    pd.testing.assert_series_equal(result, expected_result)


def test_add_string_column_from_variable(intermediate_parser_instance):
    trips_data = pd.DataFrame({
        "var1dataset1": ["1", "2", "2", "3"],
        "var2dataset1": ["A", "B", "A", "B"],
    })

    intermediate_parser_instance .trips = trips_data
    intermediate_parser_instance._add_string_column_from_variable("new_var", "var1dataset1")

    expected_result = pd.Series(["One", "Two", "Two", "Three"], name="new_var")
    pd.testing.assert_series_equal(intermediate_parser_instance.trips["new_var"], expected_result)


def test_compose_timestamp():
    data = pd.DataFrame({
        "col_year": [2023, 2023, 2023],
        "col_week": [37, 37, 38],
        "col_day": [2, 3, 4],
        "col_hour": [8, 9, 10],
        "col_min": [0, 15, 30],
    })
    col_name = "composed_timestamp"

    result = IntermediateParsing._compose_timestamp(data, "col_year", "col_week", "col_day", "col_hour", "col_min", col_name)
    expected_result = pd.DataFrame({
        "col_year": [2023, 2023, 2023],
        "col_week": [37, 37, 38],
        "col_day": [2, 3, 4],
        "col_hour": [8, 9, 10],
        "col_min": [0, 15, 30],
        "composed_timestamp": pd.DatetimeIndex(["2023-09-19 08:00:00", "2023-09-20 09:15:00", "2023-09-28 10:30:00"])
        })
    assert (result[col_name] == expected_result[col_name]).all()


def test_update_end_timestamp():
    trips = pd.DataFrame({
        "trip_end_next_day": [0, 1, 0],
        "timestamp_end": pd.DatetimeIndex(["2023-09-12 08:00:00", "2023-09-13 01:00:00", "2023-09-13 10:00:00"])
    })

    result = IntermediateParsing._update_end_timestamp(trips=trips)

    expected_result = pd.DataFrame({"timestamp_end": pd.to_datetime(["2023-09-12 08:00:00", "2023-09-14 01:00:00", "2023-09-13 10:00:00"])})
    pd.testing.assert_series_equal(result["timestamp_end"], expected_result["timestamp_end"])


def test_harmonise_variables_unique_id_names(intermediate_parser_instance):
    trips = pd.DataFrame({
        "id01": [1, 2, 3],
    })

    intermediate_parser_instance.trips = trips
    intermediate_parser_instance._harmonise_variables_unique_id_names()

    expected_result = pd.Series([1, 2, 3], name="unique_id", dtype="int")
    pd.testing.assert_series_equal(intermediate_parser_instance.trips["unique_id"], expected_result)


def test_subset_vehicle_segment(intermediate_parser_instance):
    trips = pd.DataFrame({
        "vehicle_segment_string": ["Car", "SUV", "Car", "SUV"],
        "unique_id": [1, 2, 3, 4],
    })

    intermediate_parser_instance.trips = trips
    intermediate_parser_instance._subset_vehicle_segment()

    expected_result = pd.DataFrame({
        "vehicle_segment_string": ["Car", "Car"],
        "unique_id": [1, 3],
    })
    pd.testing.assert_frame_equal(intermediate_parser_instance.trips, expected_result)



__maintainer__ = "Fabia Miorelli"
__license__ = "BSD-3-Clause"

import pytest

import pandas as pd

from ...vencopy.core.profileaggregators import ProfileAggregator

# NOT TESTED: aggregate_profiles()


@pytest.fixture
def sample_configs():
    configs = {
        'user_config': {
            'global': {
                'dataset': "dataset1",
                'debug': False,
                'absolute_path': {
                    'dataset1': '/path/to/dataset1',
                    'dataset2': '/path/to/dataset2'
                    }
                },
            'profileaggregators': {
                'aggregation_timespan': 'weekly',
                'weight_flow_profiles': False,
                'alpha': 10
            }
            },
        'dev_config': {
            'dataparsers': {
                'data_variables': False
                }
            }
        }
    return configs


@pytest.fixture
def sample_activities():
    data = {}
    activities = pd.DataFrame({
        "trip_id": [1, 2, 1, 2],
        "unique_id": [1, 1, 2, 2],
        "activity_duration": [pd.Timedelta(minutes=76), pd.Timedelta(minutes=80), pd.Timedelta(minutes=30), pd.Timedelta(minutes=75)],
        "timestamp_start": pd.DatetimeIndex(["2023-09-12 08:00:00", "2023-09-12 10:30:00", "2023-09-15 13:30:00", "2023-09-15 14:30:00"]),
        "timestamp_end": pd.DatetimeIndex(["2023-09-12 09:16:00", "2023-09-12 11:50:00", "2023-09-15 14:00:00", "2023-09-15 15:45:00"])
        })
    data['activities'] = activities
    return data


@pytest.fixture
def sample_profiles():
    profiles = pd.DataFrame({
        "drain": {'unique_id': [0, 1], 'bin': [0, 1]},
        "charging_power": {'unique_id': [0, 1], 'bin': [0, 1]},
        "uncontrolled_charging": {'unique_id': [0, 1], 'bin': [0, 1]},
        "max_battery_level": {'unique_id': [0, 1], 'bin': [0, 1]},
        "min_battery_level": {'unique_id': [0, 1], 'bin': [0, 1]}
        })
    return profiles


def test_profile_aggregator_init(sample_configs, sample_activities, sample_profiles):
    profile_aggregator = ProfileAggregator(
        configs=sample_configs,
        data=sample_activities,
        profiles=sample_profiles
    )

    assert profile_aggregator.user_config == sample_configs["user_config"]
    assert profile_aggregator.dev_config == sample_configs["dev_config"]
    assert profile_aggregator.dataset == "dataset1"
    assert profile_aggregator.weighted == False
    assert isinstance(profile_aggregator.data["activities"], pd.DataFrame)
    assert isinstance(profile_aggregator.profiles, pd.DataFrame)

    #TODO: fix test below, profiles are wrong type in fixtures
    # assert isinstance(profile_aggregator.drain, pd.DataFrame)
    # assert isinstance(profile_aggregator.charging_power, pd.DataFrame)
    # assert isinstance(profile_aggregator.uncontrolled_charging, pd.DataFrame)
    # assert isinstance(profile_aggregator.max_battery_level, pd.DataFrame)
    # assert isinstance(profile_aggregator.min_battery_level, pd.DataFrame)

    assert profile_aggregator.drain_weekly.empty
    assert profile_aggregator.charging_power_weekly.empty
    assert profile_aggregator.uncontrolled_charging_weekly.empty
    assert profile_aggregator.max_battery_level_weekly.empty
    assert profile_aggregator.min_battery_level_weekly.empty
    # TODO: add check for aggregator class instantiation



from pathlib import Path

import numpy as np
import pandas as pd
import yaml

def getData(globalConfig: dict, localPathConfig: dict, datasetID: str = 'MiD17'):
    '''
    :param globalConfig:
    :param localPathConfig:
    :param datasetID:
    :return:
    '''
    rawDataPath = Path(localPathConfig['pathAbsolute'][datasetID]) / globalConfig['files'][datasetID]['tripsDataRaw']
    logging.info(f"Starting to retrieve local data file from {rawDataPath}")
    rawData = pd.read_stata(rawDataPath, convert_categoricals=False, convert_dates=False, preserve_dtypes=False)
    dataSample = rawData.copy().loc[:, 'HP_ID_Reg'].sample(n=500, random_state=0).sort_values()
    # data= pd.DataFrame(rawData.copy().loc[:, 'HP_ID_Reg'].sample(n=500, random_state=0), columns= ['HP_ID_Reg']).sort_values(by=['HP_ID_Reg'])
    rawData= rawData.loc[:, :].where(rawData['HP_ID_Reg'].isin(dataSample)).dropna()
    return rawData


def selectColumns(rawData: pd.DataFrame):
    data = rawData.loc[:, compileVariableList(parseConfig=parseConfig)]
    return data


def compileVariableList(parseConfig, datasetID: str = 'MiD17') -> list:
    listIndex = parseConfig['dataVariables']['datasetID'].index(datasetID)
    variables = [val[listIndex] if not val[listIndex] == 'NA' else 'NA' for key, val in parseConfig['dataVariables'].items()]
    variables.remove(datasetID)
    removeNA(variables)
    return variables


def removeNA(variables: list):
    vars = [iVar.upper() for iVar in variables]
    counter = 0
    for idx, iVar in enumerate(vars):
        if iVar == 'NA':
            del variables[idx - counter]
            counter += 1


def harmonizeVariables(data, datasetID: str = 'MiD17'):
    replacementDict = createReplacementDict(datasetID, parseConfig['dataVariables'])
    dataFlipped = {val: key for key, val in replacementDict.items()}
    dataHarmonized = data.rename(columns=dataFlipped)
    return dataHarmonized


def createReplacementDict(datasetID : str, dictRaw : dict) -> None:

    if datasetID in dictRaw['datasetID']:
        listIndex = dictRaw['datasetID'].index(datasetID)
        return {val[listIndex]: key for (key, val) in dictRaw.items()}
    else:
        raise ValueError(f'Data set {datasetID} not specified in MiD variable dictionary.')


def replaceHouseholdPersonID(dataConverted):
    dataSampled= dataConverted.drop(['H_ID_Reg', 'P_ID'], axis=1)
    np.random.seed(1)
    nums_hhPersonID = np.random.choice(range(len(dataSampled)), size=dataSampled['HP_ID_Reg'].nunique(), replace=False)
    dataSampled['HP_ID_Reg'] = dataSampled['HP_ID_Reg'].map(dict(zip(dataSampled['HP_ID_Reg'].unique(), nums_hhPersonID)))
    dataSampled.reset_index(drop=True, inplace=True)
    dataSampled.to_csv('MiD17.csv')
    logging.info('Data sampling completed')
    return dataSampled


if __name__ == '__main__':
    pathLocalPathConfig = Path.cwd().parent.parent / 'config' / 'localPathConfig.yaml'
    with open(pathLocalPathConfig) as ipf:
        localPathConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    pathParseConfig = Path.cwd().parent.parent / 'config' / 'parseConfig.yaml'
    with open(pathParseConfig) as ipf:
        parseConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    pathGlobalConfig = Path.cwd().parent.parent / 'config' / 'globalConfig.yaml'
    with open(pathGlobalConfig) as ipf:
        globalConfig = yaml.load(ipf, Loader=yaml.SafeLoader)
    rawData = getData(globalConfig=globalConfig, localPathConfig=localPathConfig)
    data = selectColumns(rawData)
    dataHarmonized = harmonizeVariables(data)
    dataSampled = replaceHouseholdPersonID(dataHarmonized)



# %%
# # Tutorial 3 venco.py
# This tutorial aims to give a more in depth overview into the GridModeler class
# and showcases some features that can be customised.

from pathlib import Path
import matplotlib.pyplot as plt

from vencopy.core.dataparsers import parse_data
from vencopy.core.flexestimators import FlexEstimator
from vencopy.core.gridmodellers import GridModeller
from vencopy.core.diarybuilders import DiaryBuilder
from vencopy.core.profileaggregators import ProfileAggregator
from vencopy.utils.utils import load_configs, create_output_folders

# %%
base_path = Path.cwd() / 'vencopy'
configs = load_configs(base_path)
create_output_folders(configs=configs)

# Adapt relative paths in config for tutorials
configs['dev_config']['global']['relative_path']['parse_output'] = Path.cwd() / "vencopy" / configs['dev_config']['global']['relative_path']['parse_output']
configs['dev_config']['global']['relative_path']['diary_output'] = Path.cwd() / "vencopy" / configs['dev_config']['global']['relative_path']['diary_output']
configs['dev_config']['global']['relative_path']['grid_output'] = Path.cwd() / "vencopy" / configs['dev_config']['global']['relative_path']['grid_output']
configs['dev_config']['global']['relative_path']['flex_output'] = Path.cwd() / "vencopy" / configs['dev_config']['global']['relative_path']['flex_output']
configs['dev_config']['global']['relative_path']['aggregator_output'] = Path.cwd() / "vencopy" / configs['dev_config']['global']['relative_path']['aggregator_output']
configs['dev_config']['global']['relative_path']['processor_output'] = Path.cwd() / "vencopy" / configs['dev_config']['global']['relative_path']['processor_output']

# Set reference dataset
dataset_id = 'MiD17'

# Modify the config file to point to the .csv file in the sampling folder in the
# tutorials directory where the dataset for the tutorials lies.
configs["user_config"]["global"]["absolute_path"]["vencopy_root"] = Path.cwd()
configs['user_config']['global']['absolute_path'][dataset_id] = Path.cwd() / 'tutorials' / 'data_sampling'

# Similarly we modify the dataset_id in the global config file
configs['dev_config']['global']['files'][dataset_id]['trips_data_raw'] = dataset_id + '.csv'

# We also modify the config file for dataparsers by removing some of the columns
# that are normally parsed from the MiD, which are not available in our
# semplified test dataframe
del configs['dev_config']['dataparsers']['data_variables']['household_id']
del configs['dev_config']['dataparsers']['data_variables']['person_id']

# %%
configs['dev_config']['global']['relative_path']['parse_output']

# %%
# ## GridModeller config file

# Let's print the GridModeller config file.

configs['user_config']['gridmodellers']

# %% As we can see the GridModeler config file contains two keys:
# chargingInfrastructureMappings and chargingInfrastructureDistributions. The
# first one basically sets for which trip purpose the infrastructure
# availability should be considered, the second one specifies the probabilities
# given for each location (trip purpose) and the respective charging power.

# ## _GridModeller_ class

# The charging infrastructure allocation makes use of a basic charging
# infrastructure model, which assumes the availability of charging stations when
# vehicles are parked. Since the analytical focus of the framework lies on a
# regional level (NUTS1-NUTS0), the infrastructure model is kept simple in the
# current version. Charging availability is allocated based on a binary
# True–False mapping to a respective trip purpose in the venco.py config. Thus,
# different scenarios describing different charging availability scenarios,
# e.g., at home or at home and at work etc. can be distinguished, but neither a
# regional differentiation nor a charging availability probability or
# distribution are assumed.

# At the end of the execution of the GridModeler class, the available charging
# power during parking times is added to the activities dataframe.

# %%
# Run the first two classes to generate data

data = parse_data(configs=configs)
data = data.process()

grid = GridModeller(configs=configs, activities=data)
grid.assign_grid()

# %%
# Estimate charging flexibility based on driving profiles and charge
# connection
flex = FlexEstimator(configs=configs, activities=grid.activities)
flex.estimate_technical_flexibility_through_iteration()

diary = DiaryBuilder(configs=configs, activities=flex.activities)
diary.create_diaries()

profiles = ProfileAggregator(configs=configs, activities=diary.activities, profiles=diary)
profiles.aggregate_profiles()

plt.figure(figsize=(10, 6))
plt.plot(profiles.uncontrolled_charging_weekly.index, profiles.uncontrolled_charging_weekly.iloc[:])
plt.xlim(profiles.uncontrolled_charging_weekly.index[0], profiles.uncontrolled_charging_weekly.index[-1])
plt.title('Uncontrolled Charge Volume')
plt.xlabel('Time')
plt.ylabel('Charge Volume')
plt.grid(True)
plt.show()

# %%
# We can now change the grid availability from home to the
# workplace only and see how this affects the available charging flexibility.

configs['user_config']['gridmodellers']['charging_infrastructure_mappings']['HOME'] = False
configs['user_config']['gridmodellers']['charging_infrastructure_mappings']['WORK'] = True
configs['user_config']['gridmodellers']['charging_infrastructure_mappings']

# %%
grid = GridModeller(configs=configs, activities=data)
grid.assign_grid()

# %%
# Estimate charging flexibility based on driving profiles and charge
# connection
flex = FlexEstimator(configs=configs, activities=grid.activities)
flex.estimate_technical_flexibility_through_iteration()

diary = DiaryBuilder(configs=configs, activities=flex.activities)
diary.create_diaries()

profiles = ProfileAggregator(configs=configs, activities=diary.activities, profiles=diary)
profiles.aggregate_profiles()

plt.figure(figsize=(10, 6))
plt.plot(profiles.charging_power_weekly.index, profiles.charging_power_weekly.iloc[:])
plt.xlim(profiles.charging_power_weekly.index[0], profiles.charging_power_weekly.index[-1])
plt.title('Uncontrolled Charge Volume')
plt.xlabel('Time')
plt.ylabel('Charge Volume')
plt.grid(True)
plt.show()

# %%
# From the plots we can see how different the charging power
# looks, when the option of home charging is not available.

# ## Next Steps

# In the next tutorial, you will learn more in detail the internal workings of
# the FlexEstimator class and how to customise some settings.



__maintainer__ = "Fabia Miorelli"
__license__ = "BSD-3-Clause"

import pytest

import pandas as pd
import numpy as np

from ....vencopy.core.dataparsers.parkinference import ParkInference
from ....vencopy.core.dataparsers.parkinference import OvernightSplitter

# NOT TESTED: add_parking_rows(), __adjust_park_timestamps(), __overnight_split_decider()


@pytest.fixture
def sample_configs():
    configs = {
        'user_config': {
            'global': {
                'dataset': "dataset1",
                'debug': False,
                'absolute_path': {
                    'dataset1': '/path/to/dataset1',
                    'dataset2': '/path/to/dataset2'
                    },
                'consider_temperature_cycle_dependency': {
                    'annual': True,
                    'season': 'all',
                    'daily': False
                } 
                },
            'dataparsers': {
                "location_park_before_first_trip": {
                    "dataset1": "HOME",
                    "dataset2": "HOME"}
            },
            'diarybuilders': {
                'time_resolution': 15
            }
            },
        'dev_config': {
            'dataparsers': {
                'data_variables': False
                }
            }
        }
    return configs


@pytest.fixture
def sample_activities():
    activities = pd.DataFrame({
        "trip_id": [1, 2, 1, 2],
        "unique_id": [1, 1, 2, 2],
        "activity_duration": [pd.Timedelta(minutes=76), pd.Timedelta(minutes=80), pd.Timedelta(minutes=30), pd.Timedelta(minutes=75)],
        "timestamp_start": pd.DatetimeIndex(["2023-09-12 08:00:00", "2023-09-12 10:30:00", "2023-09-15 13:30:00", "2023-09-15 14:30:00"]),
        "timestamp_end": pd.DatetimeIndex(["2023-09-12 09:16:00", "2023-09-12 11:50:00", "2023-09-15 14:00:00", "2023-09-15 15:45:00"])
        })
    return activities


def test_park_inference_init(sample_configs):
    park_inference = ParkInference(sample_configs)

    assert park_inference.user_config == sample_configs["user_config"]
    assert park_inference.activities_raw is None
    assert isinstance(park_inference.overnight_splitter, OvernightSplitter)


def test_copy_rows(sample_activities):
    result = ParkInference._copy_rows(sample_activities)
    result = result[["trip_id", "park_id"]]

    expected_result = pd.DataFrame({
        "trip_id": [1, np.nan, 2, np.nan, 1, np.nan, 2, np.nan],
        "park_id": [np.nan, 1, np.nan, 2, np.nan, 1, np.nan, 2],
    })

    assert len(result) == 2 * len(sample_activities)
    assert result.equals(expected_result)


def test_add_util_attributes(sample_activities):
    result = ParkInference._add_util_attributes(sample_activities)

    assert "previous_unique_id" in result.columns
    assert "is_first_activity" in result.columns
    assert "next_unique_id" in result.columns
    assert "is_last_activity" in result.columns

    assert result["previous_unique_id"].equals(pd.Series([0, 1, 1, 2]))
    assert result["is_first_activity"].equals(pd.Series([True, False, True, False]))
    assert result["next_unique_id"].equals(pd.Series([1, 2, 2, 0]))
    assert result["is_last_activity"].equals(pd.Series([False, True, False, True]))


def test_add_park_act_before_first_trip(sample_activities, sample_configs):
    input_data = ParkInference._add_util_attributes(sample_activities)
    result = ParkInference._add_park_act_before_first_trip(input_data, user_config=sample_configs["user_config"])

    assert "park_id" in result.columns
    assert "purpose_string" in result.columns

    assert result.loc[result["is_first_activity"], "park_id"].equals(pd.Series([float(0), float(0)], index=[0, 2]))
    assert result.loc[result["is_first_activity"], "purpose_string"].equals(pd.Series(["HOME", "HOME"], index=[0, 2]))
    assert result.loc[result["is_first_activity"] & (result["park_id"] == 0), "trip_id"].equals(pd.Series([np.nan, np.nan], index=[0, 2]))


def test_adjust_park_attrs(sample_activities, sample_configs):
    input_data = ParkInference._add_util_attributes(sample_activities)
    input_data = ParkInference._add_park_act_before_first_trip(input_data, user_config=sample_configs["user_config"])
    result = ParkInference._adjust_park_attrs(input_data)

    assert "trip_distance" in result.columns
    assert "travel_time" in result.columns
    assert "trip_is_intermodal" in result.columns
    assert "column_from_index" in result.columns

    parking_activities = result[result["trip_id"].isna()]
    assert all(parking_activities["trip_distance"].isna())
    assert all(parking_activities["travel_time"].isna())
    assert all(parking_activities["trip_is_intermodal"].isna())


    expected_column_from_index = [0, 0, 1, 2, 2, 3]
    assert result["column_from_index"].equals(pd.Series(expected_column_from_index, index=[0, 0, 1, 2, 2, 3]))


def test_drop_redundant_columns(sample_activities, sample_configs):
    input_data = ParkInference._add_util_attributes(sample_activities)
    input_data = ParkInference._add_park_act_before_first_trip(input_data, user_config=sample_configs["user_config"])
    input_data = ParkInference._adjust_park_attrs(input_data)
    input_data["trip_start_clock"] = pd.NA
    input_data["trip_start_year"] = pd.NA
    input_data["trip_start_month"] = pd.NA
    input_data["trip_start_week"] = pd.NA
    input_data["trip_start_hour"] = pd.NA
    input_data["trip_start_minute"] = pd.NA
    input_data["trip_end_clock"] = pd.NA
    input_data["trip_end_hour"] = pd.NA
    input_data["trip_end_minute"] = pd.NA
    result = ParkInference._drop_redundant_columns(input_data)

    expected_columns = [
        "trip_id",
        "unique_id",
        "activity_duration",
        "timestamp_start",
        "timestamp_end",
        "park_id",
        "is_first_activity",
        "purpose_string",
    ]

    assert all(col in result.columns for col in expected_columns)

    dropped_columns = [
        "trip_start_clock",
        "trip_end_clock",
        "trip_start_year",
        "trip_start_month",
        "trip_start_week",
        #"trip_start_hour",
        "trip_start_minute",
        "trip_end_hour",
        "trip_end_minute",
        "previous_unique_id",
        "next_unique_id",
        "column_from_index",
    ]
    assert not any(col in result.columns for col in dropped_columns)



__maintainer__ = "Fabia Miorelli"
__license__ = "BSD-3-Clause"

import pytest
import pandas as pd

from dateutil import parser
from pathlib import Path

from ....vencopy.core.dataparsers.dataparsers import DataParser

# NOT TESTED: _load_data(), _load_unencrypted_data(), _load_encrypted_data(), _harmonise_variables(), _filter(), _simple_filters(), _complex_filters(), write_output(), process()


@pytest.fixture
def sample_configs():
    configs = {
        'user_config': {
            'global': {
                'debug': False,
                'number_lines_debug': 5000,
                'absolute_path': {
                    'dataset1': '/path/to/dataset1',
                    'dataset2': '/path/to/dataset2'
                    },
                'consider_temperature_cycle_dependency': {
                    'annual': True,
                    'season': 'all',
                    'daily': False
                }    
                },            
            'dataparsers': {
                'seasons': {
                    'consider_season': True,
                    'season': "winter"
                    }
                }
            },
            
        'dev_config': {
            'dataparsers': {
                'data_variables': {
                    'dataset': ['dataset1', 'dataset2', 'dataset3']
                    }},
                'global': {
                    'files': {
                        "dataset1": {
                            "trips_data_raw": "trips01.csv"
                        },
                        "dataset2": {
                            "trips_data_raw": "trips02.csv"
                        }
                    }
                }
            }
        }
    return configs


def test_dataparser_init(sample_configs):
    dataset = "dataset1"
    parser = DataParser(sample_configs, dataset)

    assert parser.user_config == sample_configs["user_config"]
    assert parser.dev_config == sample_configs["dev_config"]
    assert parser.debug == sample_configs["user_config"]["global"]["debug"]
#   assert parser.season == sample_configs["user_config"]["dataparsers"]["season"]
    assert parser.dataset == dataset
    assert parser.raw_data_path == Path("/path/to/dataset1/trips01.csv")
    assert parser.raw_data is None
    assert parser.trips is None
    assert parser.activities is None
    assert parser.filters == {}


def test_check_dataset_id(sample_configs):
    dataset = 'dataset2'
    mock_data_parser = DataParser(configs=sample_configs, dataset=dataset)
    result = mock_data_parser._check_dataset_id(dataset=dataset)
    assert result == dataset

    dataset = "non_existent_dataset"
    with pytest.raises(AssertionError) as excinfo:
        mock_data_parser._check_dataset_id(dataset=dataset)
    assert "Defined dataset non_existent_dataset not specified" in str(excinfo.value)

    dataset = "invalid_dataset"
    with pytest.raises(AssertionError) as e:
        mock_data_parser._check_dataset_id(dataset=dataset)

    expected_error_message = (f"Defined dataset {dataset} not specified under "
                              "data_variables in dev_config. Specified "
                              "dataset_ids are ['dataset1', 'dataset2', 'dataset3'].")
    assert str(e.value) == expected_error_message


def test_create_replacement_dict():
    data_variables = {"dataset": ["dataset1", "dataset2"],
                      "vencopy_var1": ["dataset1_var1", "dataset2_var1"],
                      "vencopy_var2": ["dataset1_var2", "dataset2_var2"],}

    dataset = "dataset1"
    expected_result = {"dataset1": "dataset", "dataset1_var1": "vencopy_var1", "dataset1_var2": "vencopy_var2"}
    result = DataParser._create_replacement_dict(dataset=dataset, data_variables=data_variables)
    assert expected_result == result

    dataset = "dataset3"
    # Use a context manager to check for the raised ValueError
    with pytest.raises(ValueError, match="Dataset dataset3 not specified in dev_config variable dictionary."):
        DataParser._create_replacement_dict(dataset=dataset, data_variables=data_variables)


def test_check_filter_dict():
    valid_dict = {
        "filter1": [1, 2, 3],
        "filter2": ["A", "B", "C"],
        "filter3": [True, False, True]
    }
    DataParser._check_filter_dict(valid_dict) 

    invalid_dict = {
        "filter1": [1, 2, 3],
        "filter2": "A", 
        "filter3": [True, False, True]
    }
    with pytest.raises(AssertionError) as excinfo:
        DataParser._check_filter_dict(invalid_dict)
    assert "Not all values in filter dictionaries are lists." in str(excinfo.value)


@pytest.fixture
def sample_data_frame_filters():
    data = {"transport_mode": ["car", "bus", "bike"],
            "age": [25, 30, 40]
            }
    return pd.DataFrame(data)


def test_set_include_filter(sample_data_frame_filters):
    include_filter_dict = {
        "transport_mode": ["car"],
        "age": [30, 35],
    }

    result = DataParser._set_include_filter(dataset=sample_data_frame_filters, include_filter_dict=include_filter_dict)

    assert isinstance(result, pd.DataFrame)
    assert set(result.columns) == set(include_filter_dict.keys())

    expected_result = pd.DataFrame({
        "transport_mode": [True, False, False],
        "age": [False, True, False],
    }, index=sample_data_frame_filters.index)
    assert result.equals(expected_result)


def test_set_exclude_filter(sample_data_frame_filters):
    exclude_filter_dict = {
        "transport_mode": ["car"],
        "age": [30, 35],
    }

    result = DataParser._set_exclude_filter(dataset=sample_data_frame_filters, exclude_filter_dict=exclude_filter_dict)

    assert isinstance(result, pd.DataFrame)
    assert set(result.columns) == set(exclude_filter_dict.keys())

    expected_result = pd.DataFrame({
        "transport_mode": [False, True, True],
        "age": [True, False, True],
    }, index=sample_data_frame_filters.index)
    assert result.equals(expected_result)


def test_set_greater_than_filter(sample_data_frame_filters):
    greater_than_filter_dict = {
        "age": [30],
    }

    result = DataParser._set_greater_than_filter(dataset=sample_data_frame_filters, greater_than_filter_dict=greater_than_filter_dict)

    assert isinstance(result, pd.DataFrame)
    assert set(result.columns) == set(greater_than_filter_dict.keys())

    expected_result = pd.DataFrame({
        "age": [False, True, True],
    }, index=sample_data_frame_filters.index)
    assert result.equals(expected_result)


def test_set_smaller_than_filter(sample_data_frame_filters):
    smaller_than_filter_dict = {
        "age": [25],
    }

    result = DataParser._set_smaller_than_filter(dataset=sample_data_frame_filters, smaller_than_filter_dict=smaller_than_filter_dict)

    assert isinstance(result, pd.DataFrame)
    assert set(result.columns) == set(smaller_than_filter_dict.keys())

    expected_result = pd.DataFrame({
        "age": [True, False, False],
    }, index=sample_data_frame_filters.index)
    assert result.equals(expected_result)


@pytest.fixture
def sample_data_frame_other_filters():
    data = {
        "unique_id": [1, 1, 2, 2, 3],
        "timestamp_start": ["2023-09-01 08:00", "2023-09-01 08:45", "2023-09-01 10:00", "2023-09-01 10:30", "2023-09-01 11:00"],
        "timestamp_end": ["2023-09-01 09:00", "2023-09-01 09:30", "2023-09-01 10:15", "2023-09-01 11:15", "2023-09-01 11:30"],
        "trip_distance": [60.0, 15.0, 10.0, 10.0, 10.0],
        "travel_time": [60, 45, 90, 90, 90],
    }
    data["timestamp_start"] = [parser.parse(x) for x in data["timestamp_start"]]
    data["timestamp_end"] = [parser.parse(x) for x in data["timestamp_end"]]
    return pd.DataFrame(data)


def test_filter_inconsistent_speeds(sample_data_frame_other_filters):
    lower_speed_threshold = 10.0
    higher_speed_threshold = 30.0
    result = DataParser._filter_inconsistent_speeds(sample_data_frame_other_filters, lower_speed_threshold, higher_speed_threshold)

    assert isinstance(result, pd.Series)
    assert len(result) == len(sample_data_frame_other_filters)

    expected_result = pd.Series([False, True, False, False, False])
    assert result.equals(expected_result)


def test_filter_inconsistent_travel_times(sample_data_frame_other_filters):
    result = DataParser._filter_inconsistent_travel_times(sample_data_frame_other_filters)

    assert isinstance(result, pd.Series)
    assert len(result) == len(sample_data_frame_other_filters)

    expected_result = pd.Series([True, True, False, False, False])
    assert result.equals(expected_result)


def test_filter_overlapping_trips(sample_data_frame_other_filters):
    result = DataParser._filter_overlapping_trips(sample_data_frame_other_filters, lookahead_periods=1)

    assert isinstance(result, pd.Series)
    assert len(result) == len(sample_data_frame_other_filters)

    expected_result = pd.Series([False, False, True, True, True])
    assert result.equals(expected_result)


def test_identify_overlapping_trips(sample_data_frame_other_filters):
    result = DataParser._identify_overlapping_trips(sample_data_frame_other_filters, 1)

    assert isinstance(result, pd.Series)
    assert len(result) == len(sample_data_frame_other_filters)

    expected_result = pd.Series([True, False, True, True, True])
    assert result.equals(expected_result)


# def test_filter_analysis(capsys):
#     filter_data = pd.DataFrame({
#         "Filter1": [True, False, True, True],
#         "Filter2": [False, True, True, False],
#     })
#
#     DataParser._filter_analysis(filter_data)
#
#     captured = capsys.readouterr()
#     expected_output = (
#         "All filters combined yielded that a total of 1 trips are taken into account.\n"
#         "This corresponds to 25.0 percent of the original data.\n"
#     )
#     assert captured.out == expected_output



__maintainer__ = "Fabia Miorelli"
__license__ = "BSD-3-Clause"


import pytest

import pandas as pd
from pathlib import Path

from ...vencopy.core.gridmodellers import GridModeller


# TESTS GridModeller class instantiation
# def test_gridmodeler_instantiation():
#     configs = {
#         "user_config": {
#             "global": {
#                 "dataset": "sample_dataset",
#             },
#             "gridmodelers": {
#                 "grid_model": "sample_grid_model",
#                 "force_last_trip_home": True,
#                 "charging_infrastructure_mappings": {
#                     "location1": True,
#                     "location2": False,
#                 },
#                 # TODO: check if syntax reflects configs
#                 # "grid_availability_distribution":
#                 #     "location1": [[11, 0.2], [22, 0.2], [0, 0.6]],
#                 #     "location2": [[11, 0.2], [22, 0.2], [0, 0.6]],
#                 },
#             },
#         "dev_config": {},
#     }
#     # TODO: fix activities dataset
#     activities = ["activity1", "activity2"]

#     my_obj = GridModeller(configs, activities)

#     assert isinstance(my_obj, GridModeller)

#     assert my_obj.user_config == configs["user_config"]
#     assert my_obj.dev_config == configs["dev_config"]
#     assert my_obj.dataset == "sample_dataset"
#     assert my_obj.grid_model == "sample_grid_model"
#     assert my_obj.activities == activities
#     assert my_obj.grid_availability_simple == {
#         "location1": True,
#         "location2": False,
#     }
#     assert my_obj.grid_availability_probability == {
#                     "location1": [0.1, 0.9],
#                     "location2": [0.1, 0.9],
#                 }
#     assert my_obj.charging_availability is None


# TESTS _adjust_power_short_parking_time
@pytest.fixture
def sample_activities():
    data = {
        "park_id": [1, 2, 3],
        "time_delta": [300, 600, 1200],  # 5 minutes, 10 minutes, 20 minutes
        "rated_power": [3.6, 11, 22],
    }
    return pd.DataFrame(data)

# def test_adjust_power_short_parking_time(sample_activities: pd.DataFrame):
#     user_config = {
#         "gridmodelers": {
#             "minimum_parking_time": 900,  # 15 minutes in seconds
#         }
#     }

#     charging_instance = GridModeller(user_config, sample_activities.copy())
#     charging_instance.__adjust_power_short_parking_time()

#     expected_rated_power = [0, 0, 30]  # Only the last activity should have rated_power unchanged
#     assert list(charging_instance.activities["rated_power"]) == expected_rated_power
